
/* SSD1306Wire.h functions, see https://github.com/ThingPulse/esp8266-oled-ssd1306

  display.clear()        - clear the local pixel buffer
  display.display()      - write the buffer to the display
  display.displayOff()   - turn the display OFF
  display.displayOn()    - turn the display ON
  display.drawCircle(int16_t x, int16_t y, int16_t radius)
  display.drawFastImage(int16_t x, int16_t y, int16_t width, int16_t height, const uint8_t *image)
  display.drawHorizontalLine(int16_t x, int16_t y, int16_t length)
  display.drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1)  - draw from x0k,y0 to x1,y1
  display.drawProgressBar(uint16_t x, uint16_t y, uint16_t width, uint16_t height, uint8_t progress)
  display.drawRect(int16_t x, int16_t y, int16_t width, int16_t height) - draw a rectangle
  display.drawString(int16_t x, int16_t y, String text)
  display.drawStringMaxWidth(int16_t x, int16_t y, int16_t maxLineWidth, String text)
  display.drawVerticalLine(int16_t x, int16_t y, int16_t length)
  display.drawXbm(int16_t x, int16_t y, int16_t width, int16_t height, const char* xbm)
  display.fillCircle(int16_t x, int16_t y, int16_t radius)
  display.fillRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.flipScreenVertically()   - flip the current screen upside down
  disply.init()         - initialise the display
  disply.invertDisplay()- inverted display mode
  disply.end()          - free memory used by the display
  disply.displayOn()    - turn the display ON
  display.mirrorScreen() - draw the screen mirrored
  disply.normalDisplay()- normal display mode
  disply.reconnect()    - connect again through I2C
  display.setBrightness(uint8_t)
  display.setColor(OLEDDISPLAY_COLOR color) - BLACK, WHITE, INVERSE
  display.setContrast(uint8_t contrast, uint8_t precharge = 241, uint8_t comdetect = 64)
  display.setFont(const uint8_t* fontData)
  display.setPixel(int16_t x, int16_t y)  - set a pixel at x,y
  display.setTextAlignment(OLEDDISPLAY_TEXT_ALIGNMENT textAlignment)  - TEXT_ALIGN_LEFT, TEXT_ALIGN_CENTER, TEXT_ALIGN_RIGHT, TEXT_ALIGN_CENTER_BOTH

  W = display.getStringWidth(const char* text, uint16_t length)
  W = display.getStringWidth(String text)
  
  Display size is 128 x 64 (w x h) pixels
  Default fonts are ArialMT at 10, 16 & 24 pixels high
  Text options include:
    2 x 24, + 1 x 10 pnt lines
    1 x 24, + 2 x 16 pnt lines
    1 x 24, + 3 x 10 pnt lines
    4 x 16 pnt lines, no line spaces, decending characters will merge
    3 x 16, + 1 x 10 pnt lines
    6 x 10 pnt lines, no line spaces, decending characters will merge
*/

// --------------------------------------------------------------------------------

void Display_40ms() {
  // called every 40ms from the main loop
  // if DispDel > 0 then we wait for DispDel cycles to complete before update
  if (DispNOP) {return;}  // this function is aborted, no I2C activity
  
  if (DispDel > 0) {DispDel--; return;} // don't update display if DispDel > 0
  if (DispClr) {Display_clear(); DispClr = false; return;}
  if (DispDisp) {Display_display(); DispDisp = false; DispClr = true; return;}

  // we refresh individual displays based on a counter, DispCnt
//  Serial.println("DispMode=" + String(DispMode));
  DispCnt--;
  if (DispCnt < 1) {
    // DispCnt is reset for display updates, so can be different for each display mode
    DispCnt = 1;  // default, will trigger on each cycle
    if (DispBack) {
      // draw a back arrow on the screen
//      display.clear();
//      display.drawLine(48,32,80,16);
//      display.drawLine(48,32,80,48);
//      display.drawLine(80,16,80,48);
//      DispCnt = 10;
    } else {
      if (DispNow > 0) {
        // special display modes used in calibration etc.
        switch(DispNow) {
//          case 1: DispCnt = 3; Display_ServoVals(); break;
        }
      } else {
        if (TEST) {
          // in TEST mode we use a special display
          switch(DispMode) {
//            case 0: Display_Test_Battery(); break;
            case 1: Display_Test_PWM(); break;
            case 2: Display_Test_MPU(); break;
            case 3: Display_Test_VLX(); break;
            case 4: Display_Opt_Sens(); break;
          }
          DispModeCnt++; if (DispModeCnt >= 20) {
            DispModeCnt = 0; DispMode++;
            if (DispMode > 4) {DispMode = 0;}
          }
        } else {
          // normal menu screens, display depends on menu modes
          switch(DispMode) {
            case   0: Display_Sleep_Mode(); break;
            case   1: Display_Tasks(); break;
            
            case  10: Display_Drive_Mode(); break;
            case  11: Display_Motor_PWM(); break;
            case  12: Display_Slot_Periods(); break;
            case  13: Display_Slot_Counts(); break;
            case  14: Display_Slot_RPM(); break;
            case  15: Display_Joysticks(); break;
            
            case  20: Display_LTOF_Mode(); break;
            case  21: Display_Range(); break;

            case  30: Display_Pixy2_Mode(); break;
            case  31: Display_PixyCam(); break;
            case  32: Display_PixyBlocks(); break;

            case  40: Display_Select_Settings(); break;
            case  41: Display_Set(1); break;  // set camera
            case  42: Display_Set(2); break;  // set LED lights

            case  99: Display_Select_Mode(); break;
          }
        }
      }
    }
    // the following line puts a rectangle around every display, showing the boundry region
//    Border = true;
    if ((Border) || TEST) {Display_drawRect(0,0,128,64);}
  }
}

// --------------------------------------------------------------------------------

//void Display_Autonomous() {
//  // display Autonomous data
//  display.setFont(ArialMT_Plain_16);
//  display.setTextAlignment(TEXT_ALIGN_CENTER);
//  // Y64 = 2,16,6,16,6,16,2
//  display.drawString(64, 2, "Autonomous");
//  display.drawString(64,24, "Range " + String(Range));
//  switch(AutoMode) {
//    case 0: Any$ = ""; break;
//    case 1: Any$ = "Forward"; break;
//    case 2: Any$ = "Scanning"; break;
//    case 3: Any$ = "Turning"; break;
//    case 4: Any$ = "Backward"; break;
//  }
//  display.drawString(64,46,Any$);
//  display.display();  // display this immediately
//  DispClr = true;     // set display clear flag for next cycle
//  DispDel = 3;        // wait 200ms before displaying next count
//}

// --------------------------------------------------------------------------------

void Display_Drive_Mode() {
  // MainMode 1 display  
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 0,24,2,16,4,10
  Display_drawString(64, 2, "DRIVE");
  Display_drawString(64, 26, "MODE");
  Display_setFont(16);
  Snore++; if (Snore > 4) {Snore = 0;}
  switch (Snore) {
    case 0: Display_drawString(48, 49, ">"); break;
    case 1: Display_drawString(56, 49, ">"); break;
    case 2: Display_drawString(64, 49, ">"); break;
    case 3: Display_drawString(72, 49, ">"); break;
    case 4: Display_drawString(80, 49, ">"); break;
  }
  DispDisp = true;    // set display flag for next cycle
  DispDel = 3;        // wait 120ms before displaying next count
  DispNxtCnt++; if (DispNxtCnt > 10) {DispMode = 11;} // auto branch to PWM display
}

// --------------------------------------------------------------------------------

void Display_Init() {
  // initialise the OLED display
  display.init();
  display.clear();
  display.setContrast(255);
//  display.flipScreenVertically();
  display.display();
  DispClr = false; DispDisp = false;
}

// --------------------------------------------------------------------------------

void Display_Intro() {
  // first screen to be displayed
  // default fonts are ArialMT_Plain_10, ArialMT_Plain_16 and ArialMT_Plain_24
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 0,24,2,16,4,10
  Display_drawString(64, 0, "PixyBot");
  Display_drawString(64, 24, "2");
  Display_setFont(10);
  Display_drawString(64, 51, "Batt.= " + String(float(BatAvg)/BatCal) + "v  " + String(AnyLong) + "%");
  Display_display();
  DispClr = true;  // display intro, then clear memory ready for first screen
}

// --------------------------------------------------------------------------------

void Display_Joysticks() {
  // display joystick data
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 4, "Joysticks");
  // Y64 = 16,4,10,4,10
  if (!WiFiEn) {
    // problems with WiFi
    if (WiFiConnected) {
      Display_drawString( 64, 28, "WiFi Connected");
      Display_drawString( 64, 47, "NOT Enabled");
    } else {
      Display_drawString( 64, 28, "No WiFi Link");
      Display_drawString( 64, 47, "Inaccessible");
    }
  } else {
    Display_setTextAlignment(TEXT_ALIGN_LEFT);
    Display_drawString(  4, 28, "JX:" + String(JoyX));
    Any$ = "0"; if (CZ & 2) {Any$ = "1";}
    if (CZ & 1) {Any$ += "1";} else {Any$ += "0";}
    Display_drawString(  4, 47, "CZ:" + Any$);
    Display_drawString( 68, 28, "JY:" + String(JoyY));
    Display_drawString( 68, 47, " D:" + String(JoyMode));
  }
  DispDisp = true;    // set display flag for next cycle
  DispDel = 1;        // wait 40ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_LTOF_Mode() {
  // MainMode 2 display  
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 0,24,2,16,4,10
  Display_drawString(64, 2, "LTOF");
  Display_drawString(64, 26, "MODE");
  Display_setFont(16);
  Snore++; if (Snore > 6) {Snore = 0;}
  switch (Snore) {
    case 0: Display_drawString(40, 49, "|"); break;
    case 1: Display_drawString(48, 49, "|"); break;
    case 2: Display_drawString(56, 49, "|"); break;
    case 3: Display_drawString(64, 49, "|"); break;
    case 4: Display_drawString(72, 49, "|"); break;
    case 5: Display_drawString(80, 49, "|"); break;
    case 6: Display_drawString(88, 49, "|"); break;
  }
  DispDisp = true;    // set display flag for next cycle
  DispDel = 3;        // wait 120ms before displaying next count
  DispNxtCnt++; if (DispNxtCnt > 10) {DispMode = 21;} // auto branch to Range display
}

// --------------------------------------------------------------------------------

void Display_Motor_PWM() {
  // display the PWM values being applied to the motors
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 4, "Motor  PWM");
  // Y64 = 16,4,10,4,10
  Display_drawString( 64, 22, String(Steer));
  Display_drawString( 24, 28, String(PWM_FL));
  Display_drawString(103, 28, String(PWM_FR));
  Display_drawString( 24, 47, String(PWM_RL));
  Display_drawString(103, 47, String(PWM_RR));
  if (WiFiEn) {Display_drawString( 64, 41, String(Drive));}
  if (DithON) {Display_drawRect(0,24,128,40);}
  DispDisp = true;    // set display flag for next cycle
  DispDel = 1;        // wait 40ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_PixyBlocks() {
  // display detected blocks as rectangles on the display  
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  if (!PixyEn) {
    // Y64 = 28,16,20
    Display_drawString(64,28, "Pixy Disabled");
  } else {
    // Y64 = 4,16,8,16
    if ((BlockCnt < 1) || (Pixy_mW[0] < 20)) {
      // if no blocks then display text
           if (MainTask <  32) {Display_drawString(64,4, "Pixy Blocks");}
      else if (MainTask == 32) {Display_drawString(64,4, "Pixy Tracker");}
      else if (MainTask == 33) {Display_drawString(64,4, "Pixy Turner");}
      else if (MainTask == 34) {Display_drawString(64,4, "Pixy Follower");}
      else if (MainTask == 35) {Display_drawString(64,4, "Pixy Follower");}
      else if (MainTask == 36) {Display_drawString(64,4, "Pixy Goalie");}
      if (BlockCnt < 1) {Display_drawString(64,28, "No Blocks");}
      else {Display_drawString(64,28, "Far Away");}
      Display_setFont(10);
      Display_drawString(64,53,String(PixyFps) + " fps");
    } else {
      // draw up to 3 blocks max as rectangles
      Display_setFont(10);
      for (int zb = 0;zb < BlockCnt;zb++) {
        // take each block in turn, remap and draw it
//        DispCX = map(Pixy_mX[zb],0,315,0,127);
        DispCX = map(Pixy_mX[zb],0,207,0,63) + 16;
        DispCY = map(Pixy_mY[zb],0,207,0,63);
//        DispW  = map(Pixy_mW[zb],0,315,0,127);
        DispW  = map(Pixy_mW[zb],0,207,0,63);
        DispH  = map(Pixy_mH[zb],0,207,0,63);
        Display_drawRect(DispCX - (DispW/2),DispCY - (DispH/2),DispW,DispH);
        // now include signature ID
        Display_drawString(DispCX,DispCY - 5, "S" + String(Pixy_Sg[zb]));
      }
//      if (BlockAsk == 1) {
      if (MainTask >= 32) {
        // single block tracking mode
        // draw deadband limits top and sides
        Display_drawLine( 61, 0, 66, 0); // top line X Db line, @ PxDbLL = 152,PxDbUL = 164
        Display_drawLine(  0,30,  0,33); // left side Y Db line, @ PyDbLL = 100,PyDbUL = 108
        Display_drawLine(127,30,127,33); // right side Y Db line
        // show width and height at the bottom of the display
        Display_setTextAlignment(TEXT_ALIGN_CENTER);
        if (PixyFps < 55) {Display_drawString(64,53,"F:" + String(PixyFps));}
        else {Display_drawString(64,53,String(getPanAngle()) + "," + String(getTiltAngle()));}
        Display_setTextAlignment(TEXT_ALIGN_LEFT);
        Display_drawString(4,53,"W:" + String(Pixy_mW[0]));
        Display_setTextAlignment(TEXT_ALIGN_RIGHT);
        Display_drawString(123,53,"H:" + String(Pixy_mH[0]));
      }
    }
  }
//  Serial.println(String(DispCX) + "," + String(DispCY) + "," + String(DispW) + "," + String(DispH));
  DispDisp = true;    // set display flag for next cycle
  DispDel = 1;        // wait 40ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_PixyCam() {
  // display information about Pixy2 camera  
  String zS$ = "";  // puts an 's' on the end of blocks
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 4, "Pixy2 Cam");
  // Y64 = 4,16,8,16
  if (BlockCnt < 1) {
    Display_drawString(64,28, "No Blocks");
  } else {
    if (BlockCnt > 0) {zS$ = "s";}
    Display_drawString(64,28, String(BlockCnt) + " Block" + zS$);
    // read the first block data
    if (PixyEn) {
      zS$ = "S" + String(Pixy_Sg[0]) + ",";
      zS$ += String(Pixy_mX[0]) + ",";
      zS$ += String(Pixy_mY[0]) + ",";
      zS$ += String(Pixy_mW[0]) + ",";
      zS$ += String(Pixy_mH[0]);
    }
    Display_setFont(10);
    Display_drawString(64,53,zS$);
  }
  DispDisp = true;    // set display flag for next cycle
  DispDel = 2;        // wait 80ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Pixy2_Mode() {
  // MainMode 3 display  
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 0,24,2,16,4,10
  Display_drawString(64, 2, "Pixy2");
  Display_drawString(64, 26, "MODE");
  Display_setFont(16);
  Snore++; if (Snore > 4) {Snore = 0;}
  switch (Snore) {
    case 0: Display_drawString(48, 49, "V"); break;
    case 1: Display_drawString(56, 49, "V"); break;
    case 2: Display_drawString(64, 49, "V"); break;
    case 3: Display_drawString(72, 49, "V"); break;
    case 4: Display_drawString(80, 49, "V"); break;
  }
  DispDisp = true;    // set display flag for next cycle
  DispDel = 3;        // wait 120ms before displaying next count
  DispNxtCnt++; if (DispNxtCnt > 10) {DispMode = 31;} // auto branch to PixyCam
}

// --------------------------------------------------------------------------------

void Display_Question_YN(String zQ1,String zQ2) {
  // display NO or YES question screen  
  Display_clear();
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 4,16,6,16,6,16
  Display_drawString(64, 4, zQ1);
  Display_drawString(64, 25, zQ2);
  // draw the button bars
//  display.setColor(INVERSE);  // BLACK, WHITE, INVERSE
  Display_fillRect( 0,46,37,18);
  Display_fillRect(91,46,37,18);
  Display_setColor(BLACK);
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(  7, 46, "NO");
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(126, 46, "YES");
  Display_setColor(WHITE);  // BLACK, WHITE, INVERSE
  Display_display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 25;       // wait 1 sec before display next screen
}

// --------------------------------------------------------------------------------

void Display_Range() {
  // display VL53L1X range measurement
  String zSL = ""; String zSR = "";
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 2,16,6,16,6,16,2
       if (MainTask == 22) {Display_drawString(64, 2, "Back Away");}
  else if (MainTask == 23) {Display_drawString(64, 2, "Track Mode");}
  else {Display_drawString(64, 2, "Range");}
  if (RangeEn) {
    if (RangeUL) {zSL = "< "; zSR = " >";}
    if (RangeLL) {zSL = "> "; zSR = " <";}
    Display_drawString(64,24, zSL + String(RangeRaw) + " mm" + zSR);
  } else {Display_drawString(65,24, "---");}
  if (LTOFcalc) {
    Display_setTextAlignment(TEXT_ALIGN_LEFT);
    Display_drawString(0,46,String(LTOFfps) + " fps");
    
    Display_setTextAlignment(TEXT_ALIGN_RIGHT);
    if ((LTOFspd10 == 0) || RangeUL) {
      Display_drawString(127,46, "--- m/s");
    } else {
      Display_drawString(127,46, String(float(LTOFspd10)/10.0,1) + " m/s");
    }
  } else if (HeadActive) {
    // head is moving so display the angle
    Display_setTextAlignment(TEXT_ALIGN_LEFT);
    Display_drawString(0,46,String(LTOFfps) + " fps");
    
    Display_setTextAlignment(TEXT_ALIGN_RIGHT);
//    GetHeadAngle(); 
    Display_drawString(127,46,String(HeadAng) + "°");
  } else {
    if (MainTask < 22) {Display_drawString(64,46,String(LTOFfps) + " fps");}
    else {Display_setFont(10); Display_drawString(64,53,Status$);}
  }
  Display_display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 3;        // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Select_Mode() {
  // display mode select screen  
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 4,16,6,16,6,16
  Display_drawString(64, 4, "Select  Mode");
  switch (ModeSel) {
    case 0: Display_drawString(64, 25, "SLEEP"); break;
    case 1: Display_drawString(64, 25, "DRIVE"); break;
    case 2: Display_drawString(64, 25, "LTOF"); break;
    case 3: Display_drawString(64, 25, "Pixy2"); break;
    case 4: Display_drawString(64, 25, "Settings"); break;
  }
  // draw the button bars
//  display.setColor(INVERSE);  // BLACK, WHITE, INVERSE
  Display_fillRect( 0,46,37,18);
  Display_fillRect(91,46,37,18);
  Display_setColor(BLACK);
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(  7, 46, "NO");
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(126, 46, "YES");
  Display_setColor(WHITE);  // BLACK, WHITE, INVERSE
  DispDisp = true;    // set display flag for next cycle
  DispDel = 5;        // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Select_Settings() {
  // display settings select screen  
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 4,16,6,16,6,16
  Display_drawString(64, 4, "Settings");
  switch (SetSel) {
    case 0: Display_drawString(64, 25, "Camera"); break;
    case 1: Display_drawString(64, 25, "Lights"); break;
  }
  // draw the button bars
  Display_fillRect( 0,46,37,18);
  Display_fillRect(91,46,37,18);
  Display_setColor(BLACK);
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(  7, 46, "NO");
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(126, 46, "YES");
  Display_setColor(WHITE);  // BLACK, WHITE, INVERSE
  DispDisp = true;    // set display flag for next cycle
  DispDel = 5;        // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Set(int16_t zS) {
  // display settings adjust screen
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // top line
       if (zS == 1) {Display_drawString(64, 4, "Camera Lux");}
  else if (zS == 2) {Display_drawString(64, 4, "LED Lights");}
  // now display current setting
       if (zS == 1) {Display_drawString(64,25, String(DispVal));}
  // draw the button bars
  Display_fillRect( 0,46,37,18);
  Display_fillRect(91,46,37,18);
  Display_setColor(BLACK);
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(  7, 46, "<<");
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(126, 46, ">>");
  Display_setColor(WHITE);  // BLACK, WHITE, INVERSE
  DispDisp = true;    // set display flag for next cycle
  DispDel = 5;        // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Sleep_Mode() {
  // default state display  
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 0,24,2,16,4,10
  Display_drawString(64, 2, "SLEEP");
  Display_drawString(64, 26, "MODE");
  Snore++; if (Snore > 3) {Snore = 0;}
  switch (Snore) {
    case 0:
      Display_setFont(10); Display_drawString(64, 50, "."); break;
    case 1:
      Display_setFont(10); Display_drawString(64, 50, "z"); break;
    case 2:
      Display_setFont(10); Display_drawString(64, 50, "Z"); break;
    case 3:
      Display_setFont(16); Display_drawString(64, 49, "Z"); break;
  }
  Display_setBrightness(64);  // ensure display is dim
  DispDisp = true;            // set display flag for next cycle
  DispDel = 5;                // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Slot_Counts() {
  // display the slot counts from the speed sensors
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 4, "Slot  Counts");
  // Y64 = 16,8,16,4,16
  Display_drawString( 24, 28, String(IntP36Cnt));  // FL
  Display_drawString(103, 28, String(IntP34Cnt));  // FR
  Display_drawString( 24, 48, String(IntP35Cnt));  // RL
  Display_drawString(103, 48, String(IntP39Cnt));  // RR
  DispDisp = true;    // set display flag for next cycle
  DispDel = 1;        // wait 40ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Slot_Periods() {
  // display the slot periods from the speed sensors
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 4, "Slot  Periods");
  // Y64 = 16,8,16,4,16
  Display_drawString( 32, 28, String(IntPeriods[0]));  // FL
  Display_drawString( 95, 28, String(IntPeriods[1]));  // FR
  Display_drawString( 32, 48, String(IntPeriods[2]));  // RL
  Display_drawString( 95, 48, String(IntPeriods[3]));  // RR
  DispDisp = true;    // set display flag for next cycle
  DispDel = 2;        // wait 80ms before displaying next count
  if (!IntP34Ch && !IntP35Ch && !IntP36Ch && !IntP39Ch) {
    // no change in the previous period so clear displayed values
    Reset_Slots();
  }
  IntP34Ch = false; IntP35Ch = false; IntP36Ch = false; IntP39Ch = false;
}

// --------------------------------------------------------------------------------

void Display_Slot_RPM() {
  // display the slot speed sensors RPM values
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 4, "Slot  RPMs");
  // Y64 = 16,8,16,4,16
  Display_drawString( 24, 28, String(0.1 * Int36RPM,1));  // FL
  Display_drawString(101, 28, String(0.1 * Int34RPM,1));  // FR
  Display_drawString( 24, 48, String(0.1 * Int35RPM,1));  // RL
  Display_drawString(101, 48, String(0.1 * Int39RPM,1));  // RR
  DispDisp = true;    // set display flag for next cycle
  DispDel = 2;        // wait 80ms before displaying next count
  if (!IntP34Ch && !IntP35Ch && !IntP36Ch && !IntP39Ch) {
    // no change in the previous period so clear displayed values
    Reset_Slots();
  }
  IntP34Ch = false; IntP35Ch = false; IntP36Ch = false; IntP39Ch = false;
}

// --------------------------------------------------------------------------------

void Display_Tasks() {
  // display the task values
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 4, "Tasks");
  // Y64 = 16,4,10,4,10
  Display_setFont(10);
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString( 32, 27, "Main:");
  Display_drawString( 96, 27, "Sub:");
  Display_drawString( 32, 40, "Cnt:");
  Display_drawString( 96, 40, "Next:");
  Display_drawString( 32, 53, "LED:");
  Display_drawString( 96, 53, "Wait:");
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString( 34, 27, String(MainTask));
  Display_drawString( 98, 27, String(SubTask));
  Display_drawString( 34, 40, String(SubCnt));
  Display_drawString( 98, 40, String(SubNext));
  Display_drawString( 34, 53, String(LedMode));
  Display_drawString( 98, 53, String(SubWait));
  DispDisp = true;    // set display flag for next cycle
  DispDel = 1;        // wait 40ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Test_Battery() {
  // display battery voltage  
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 8, "Battery");
  // Y64 = 16,16,16,16
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(62, 40, "Batt. =");
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(66, 40, String(float(BatAvg)/BatCal) + "v");
  DispDisp = true;    // set display flag for next cycle
  DispDel = 20;        // wait 800ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Test_MPU() {
  // display servoAny
  if (!I2C_MPU) {DispModeCnt = 20; return;}
  
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 0, "MPU");
  // Y64 = 16,16,16,16
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(62, 15, "Acc X =");
  Display_drawString(62, 30, "Acc Y =");
  Display_drawString(62, 45, "Gyr Z =");
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(64, 15, String(accelerometer_data_rawX_i));
  Display_drawString(64, 30, String(accelerometer_data_rawY_i));
  Display_drawString(64, 45, String(gyro_Z_data_raw));
  DispDisp = true;    // set display flag for next cycle
  DispDel = 2;        // wait 80ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Opt_Sens() {
  // display the state of the four optical sensors  
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 2, "Opt Sensors");
  // Y64 = 16,16,16,16
  // INT_P34  speed sensor interupt pin front-right
  // INT_P35  speed sensor interupt pin rear-left
  // INT_P36  speed sensor interupt pin front-left
  // INT_P39  speed sensor interupt pin rear-right
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString( 0, 22, "F-L = " + String(digitalRead(INT_P36)));
  Display_drawString( 0, 42, "R-L = " + String(digitalRead(INT_P35)));
  Display_drawString(64, 22, "F-R = " + String(digitalRead(INT_P34)));
  Display_drawString(64, 42, "R-R = " + String(digitalRead(INT_P39)));
  DispDisp = true;    // set display flag for next cycle
  DispDel = 2;        // wait 80ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Test_PWM() {
  // display servoAny
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 6,24,4,24,6
  int16_t zY = 20;
  Display_drawString(64, zY, "PWM " + String(TEST_PWM));
  DispDisp = true;    // set display flag for next cycle
  DispDel = 2;        // wait 80ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Test_VLX() {
  // display Ltof range
  if (!I2C_LTOF) {DispModeCnt = 20; return;}
  
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64,  0, "LTOF");
  // Y64 = 24,4,24,6
  Display_setFont(16);
  Display_drawString(64, 32, "Range " + String(Range));
  DispDisp = true;    // set display flag for next cycle
  DispDel = 2;        // wait 80ms before displaying next count
}

// --------------------------------------------------------------------------------

void DisplayText1S24(String zT0) {
  // display one line of text at 24 pt
  Display_clear();
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 20,24,20
  Display_drawString(64, 20, zT0);
  Display_display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 25;       // wait 1 sec before display next screen
}

// --------------------------------------------------------------------------------

void DisplayText2(String zT0,String zT1) {
  // display two lines of text of equal height
  Display_clear();
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 14,16,4,16,14
  Display_drawString(64, 14, zT0);
  Display_drawString(64, 34, zT1);
  if (Border) {Display_drawRect(0,0,128,64); Display_drawRect(1,1,126,62);} // thick boarder
  Display_display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 25;       // wait 1 sec before display next screen
}

// --------------------------------------------------------------------------------

void DisplayText2S24(String zT0,String zT1) {
  // display two lines of text, 1st line is large
  Display_clear();
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 8,24,8,16,8
  Display_drawString(64,  8, zT0);
  Display_setFont(16);
  Display_drawString(64, 40, zT1);
  Display_display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 25;       // wait 1 sec before display next screen
}

// --------------------------------------------------------------------------------
