
// --------------------------------------------------------------------------------

void battRead() {
  // read the analogue voltage connected to the battery every 40ms
  // this frunction is no longer called due to the limitations of ADC2 under Wi-Fi
  BatVol = analogRead(BatPin);
//  BatVol-= BatDist; BatDist++;  // uncomment this line to simulate rapid battery discharge
  BatSum = BatSum + BatVol - BatAvg;
  BatAvg = BatSum/50;     // ADC is averaged over 50 readings to remove noise
  if (TEST) {Serial.println("Battery " + String(BatVol) + "\t" + String(float(BatAvg)/BatCal) + "v\t" + String(AnyLong) + "%");}
  if (USB) {return;}

  // detect critical battery condition and disable PixyBot
//  if (BatAvg <= BatCritical){
//    // critical battery voltage reached
//    Display_Battery_Low();            // display warning
//    FastLED.clear(); FastLED.show();  // turn OFF LEDs
//    VL53L1X_OFF();                    // turn OFF LTOF
//    GoToStore(50);
//    while (TgtCnt > 0) {
//      // complete the goto rest from here
//      Leg_Engine(); delay(20);
//    }
//    // and that's it
//    while (true) {yield();}
//  }
}

// ---------------------------------------------------------------------------------

void decodeKey(int zkeyVal) {
  // decodes zkeyVal and excutes commands on '.'
  // every time we receive a character set response to 10ms later, just in case
  // there is another packet behind it coming from the controller
  next40ms = millis() + 10;
  if (zkeyVal == 10) {return;}
  if (zkeyVal == 13) {return;}
  keyChar = char(zkeyVal);
  switch (keyChar) {
    case '.': doCmd(); return;                        // command terminator
    case '~': DispMon = true; DispTx = 50; return;    // connected to OLED Mirror app
    case '-': cmdSgn = -1; return;                    // numeric sign
    case '*': cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1; return; // ESC abort command
    case '0': extendCmdVal(0); return;
    case '1': extendCmdVal(1); return;
    case '2': extendCmdVal(2); return;
    case '3': extendCmdVal(3); return;
    case '4': extendCmdVal(4); return;
    case '5': extendCmdVal(5); return;
    case '6': extendCmdVal(6); return;
    case '7': extendCmdVal(7); return;
    case '8': extendCmdVal(8); return;
    case '9': extendCmdVal(9); return;
  }
  if (cmdMode == ' ') {
    // test for new Command Mode char?
    // only accept certain characters as valid commands, or ignore them
    // delay battery update by 10 seconds whilst receiving commands
    switch (keyChar) {
      // check for valid mode and convert upper-case
      case 'j': cmdMode = 'J'; break;
      case 'J': cmdMode = 'J'; break;
      case 'p': cmdMode = 'P'; break;
      case 'P': cmdMode = 'P'; break;
      case 's': cmdMode = 'S'; break;
      case 'S': cmdMode = 'S'; break;
      case 'v': cmdMode = 'V'; break;
      case 'V': cmdMode = 'V'; break;
      case 'w': cmdMode = 'W'; break;
      case 'W': cmdMode = 'W'; break;
    } cmdType = ' '; cmdVal = 0;
  } else {
    // test for Command Type char?
    cmdType = keyChar;
    switch (keyChar) {
      // check for lower-case and convert to upper-case
      case 'a': cmdType = 'A'; break;
      case 'c': cmdType = 'C'; break;
      case 'd': cmdType = 'D'; break;
      case 'h': cmdType = 'H'; break;
      case 'l': cmdType = 'L'; break;
      case 'o': cmdType = 'O'; break;
      case 'r': cmdType = 'R'; break;
      case 's': cmdType = 'S'; break;
      case 't': cmdType = 'T'; break;
      case 'u': cmdType = 'U'; break;
      case 'v': cmdType = 'V'; break;
      case 'x': cmdType = 'X'; break;
      case 'y': cmdType = 'Y'; break;
    }
  }
}

// ---------------------------------------------------------------------------------

void doCmd() {
  // a '.' has been received on the serial port so execute command if valid
  // Commands:
  // JXnn.  - set global variable JX to nn
  // JYnn.  - set global variable JY to nn
  // PTnn.  - set target wheel PWM pointer
  // SAnn.  - set target servo angle in degrees
  // SLnn.  - set target servo lower limit in microseconds
  // STnn.  - set the target servo number 0 - 16, but don't feedback value
  // SUnn.  - set target servo upper limit in microseconds
  // SVnn.  - set PWM and servo angle in microseconds

  cmdVal*= cmdSgn;  // correct cmdVal with the sign value
  
  switch (cmdMode) {
    case ' ': break;
    case 'J':
      // commands used to set global variables in association with Joystick app
      switch (cmdType) {
        case 'X': JX = cmdVal; break; // set global variable JX to cmdVal
        case 'Y': JY = cmdVal; break; // set global variable JX to cmdVal
      } break;
    case 'P':
      // keyboard/serial commands used to set target wheel PWM pointer
      switch (cmdType) {
        case 'T': // set target servo pointer in range 0 - 7
          if (cmdVal < 0) {cmdVal = ServoPnt;}
          if (cmdVal > 7) {cmdVal = ServoPnt;}
          ServoPnt = cmdVal; break;
      } break;
    case 'S':
      // keyboard/serial commands used to set servo values
      switch (cmdType) {
        case 'A': // set target servo angle 0-1000 if Pixy2 is detected
          if (ServoPnt == 8) {servoVal[0] = cmdVal;}
          if (ServoPnt == 9) {servoVal[1] = cmdVal;}
          if (I2C_Pixy) {pixy.setServos(servoVal[0],servoVal[1]);}
          break;
        case 'L': // set target servo lower limit in microseconds
          if (ServoPnt == 8) {servoMin[0] = cmdVal;}
          if (ServoPnt == 9) {servoMin[1] = cmdVal;}
          break;
        case 'T': // set Pixy2 servo pointer in range 8 - 9
          if (cmdVal < 8) {cmdVal = ServoPnt;}
          if (cmdVal > 9) {cmdVal = ServoPnt;}
          ServoPnt = cmdVal; break;
        case 'U': // set target servo upper limit in microseconds
          if (ServoPnt == 8) {servoMax[0] = cmdVal;}
          if (ServoPnt == 9) {servoMax[1] = cmdVal;}
          break;
        case 'V': // set wheel PWM values 0 - 255
          if (ServoPnt ==  0) {ledcWrite(PwmChFntLftA, cmdVal);}
          if (ServoPnt ==  1) {ledcWrite(PwmChFntLftB, cmdVal);}
          if (ServoPnt ==  2) {ledcWrite(PwmChFntRhtA, cmdVal);}
          if (ServoPnt ==  3) {ledcWrite(PwmChFntRhtB, cmdVal);}
          if (ServoPnt ==  4) {ledcWrite(PwmChBckLftA, cmdVal);}
          if (ServoPnt ==  5) {ledcWrite(PwmChBckLftB, cmdVal);}
          if (ServoPnt ==  6) {ledcWrite(PwmChBckRhtA, cmdVal);}
          if (ServoPnt ==  7) {ledcWrite(PwmChBckRhtB, cmdVal);}
          break;
      } break;
    default:
      PrintTx += "Unknown cmd: " + cmdMode + cmdType + String(cmdVal) + "\n";
      break;
  }
  // now reset the variables
  cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1;
}

// ---------------------------------------------------------------------------------

void extendCmdVal(int zVal) {
  // adds a new digit to the right-hand end of cmdVal
  cmdVal = abs(cmdVal * 10) + zVal;
}

// --------------------------------------------------------------------------------

int16_t getPanAngle() {
  // returns the Pan angle based on the current Pan servo value
  // the value should range between 30° - 150°, with centre being 90°
  // servoVal[0] low turns to the right, servoVal[0] large turns to the left
  if (servoVal[0] < servoCtr[0]) {return map(servoVal[0],servoMin[0],servoCtr[0],150,90);}
  else {return map(servoVal[0],servoCtr[0],servoMax[0],90,30);}
}

// --------------------------------------------------------------------------------

void GetRndCol(byte zB) {
  // return a random colour of strength zB
  if (zB < 2) {zB = 2;}
  switch(random(6)) {
    case 0: Col_R =   zB; Col_G =    0; Col_B =    0; break;  // red
    case 1: Col_R = zB/2; Col_G = zB/2; Col_B =    0; break;  // yellow
    case 2: Col_R =    0; Col_G =   zB; Col_B =    0; break;  // green
    case 3: Col_R =    0; Col_G = zB/2; Col_B = zB/2; break;  // cyan
    case 4: Col_R =    0; Col_G =    0; Col_B =   zB; break;  // blue
    case 5: Col_R = zB/2; Col_G =    0; Col_B = zB/2; break;  // magenta
  }
}

// --------------------------------------------------------------------------------

void getSigAp() {
  // called when SigId may be > 0 to determine the array pointer
  SigAp = 0;
  for (int zb = 0;zb < BlockCnt;zb++) {
    // take each array block in turn and check it against SigId
    if (Pixy_Sg[zb] == SigId) {SigAp = zb; break;}
  }
}

// --------------------------------------------------------------------------------

int16_t getTiltAngle() {
  // returns the Pan angle based on the current Pan servo value
  // the value should range between 30° - 150°, with centre being 90°
  // servoVal[1] low leans Pixy2 back, servoVal[1] large leans Pixy2 forwards
  // using the 3D model the leaver arm strikes the servo at -20° from vertical (90°)
  return map(servoVal[1],servoMin[1],servoCtr[1],70,90);
}

// --------------------------------------------------------------------------------

void Lamps(uint8_t zS) {
  // send I2C message to Pixy2 turning lights ON/OFF
  pixy.setLamp(zS,zS);
}

// --------------------------------------------------------------------------------

void LED_DriveBck() {
  // generates a driving backward pattern
  switch (LED_Task) {
    case 0: // create LED pattern
      FastLED.clear();
      LED[0].setRGB(3,3,0); LED[7].setRGB(3,3,0); // yellow
      LED[1].setRGB(0,0,4); LED[6].setRGB(0,0,4); // bright blue
      LED[2].setRGB(0,0,2); LED[5].setRGB(0,0,2); // mid blue
      LED[3].setRGB(0,0,1); LED[4].setRGB(0,0,1); // min blue
      LED_Run = false; LEDshow = true; LED_Task++; break;
    case 1: // rotate LED pattern
      if (LedTick) {
        LedTick = false; LED_Cnt++;
        if (LED_Cnt >= 4) {
          LEDX[0] = LED[10]; // grab the front body LED rgb values
          LED[10] = LED[ 9]; LED[11] = LED[12]; LED[ 9] = LED[ 8]; LED[12] = LED[ 7];  LED[ 8] = LED[ 0];
          LED[ 0] = LED[ 1]; LED[ 1] = LED[ 2]; LED[ 2] = LED[ 3]; LED[ 3] = LEDX[ 0];
          LED[ 7] = LED[ 6]; LED[ 6] = LED[ 5]; LED[ 5] = LED[ 4]; LED[ 4] = LEDX[ 0];
          LED_Run = false; LEDshow = true; LED_Cnt = 0;
        }
      } break;
  }
}

// --------------------------------------------------------------------------------

void LED_DriveLft() {
  // generates a driving to the left pattern
  switch (LED_Task) {
    case 0: // create LED pattern on front face
      FastLED.clear();
      LED[ 8].setRGB(3,3,0);
      LED[ 9].setRGB(0,0,4);
      LED[10].setRGB(0,0,2);
      LED[11].setRGB(0,0,1);
      LED_Run = false; LEDshow = true; LED_Task++; break;
    case 1: // rotate LED pattern
      if (LedTick) {
        LedTick = false; LED_Cnt++;
        if (LED_Cnt >= 4) {
          LED[ 0] = LED[ 9]; LED[ 1] = LED[ 8]; LED[ 2] = LED[ 8]; LED[ 3] = LED[ 9];
          LED[ 7] = LED[10]; LED[ 6] = LED[11]; LED[ 5] = LED[11]; LED[ 4] = LED[10];
          LEDX[0] = LED[ 8]; // grab the front right LED rgb values
          LED[ 8] = LED[ 9]; LED[ 9] = LED[10]; LED[10] = LED[11]; LED[11] = LED[12];
          LED[12] = LEDX[0];
          LED_Run = false; LEDshow = true; LED_Cnt = 0;
        }
      } break;
  }
}

// --------------------------------------------------------------------------------

void LED_DriveFwd() {
  // generates a driving forward pattern
  switch (LED_Task) {
    case 0: // create LED pattern
      FastLED.clear();
      LED[3].setRGB(3,3,0); LED[4].setRGB(3,3,0); // yellow
      LED[2].setRGB(0,0,4); LED[5].setRGB(0,0,4); // bright blue
      LED[1].setRGB(0,0,2); LED[6].setRGB(0,0,2); // mid blue
      LED[0].setRGB(0,0,1); LED[7].setRGB(0,0,1); // min blue
      LED_Run = false; LEDshow = true; LED_Task++; break;
    case 1: // rotate LED pattern
      if (LedTick) {
        LedTick = false; LED_Cnt++;
        if (LED_Cnt >= 4) {
          LEDX[0] = LED[3]; // grab the last body LED rgb values
          LED[3] = LED[2]; LED[2] = LED[1]; LED[1] = LED[0]; LED[0] = LED[ 8];
          LED[4] = LED[5]; LED[5] = LED[6]; LED[6] = LED[7]; LED[7] = LED[12];
          LED[8] = LED[9]; LED[12] = LED[11]; LED[9] = LED[10]; LED[11] = LED[10];
          LED[10] = LEDX[0];
          LED_Run = false; LEDshow = true; LED_Cnt = 0;
        }
      } break;
  }
}

// --------------------------------------------------------------------------------

void LED_DriveRht() {
  // generates a driving to the right pattern
  switch (LED_Task) {
    case 0: // create LED pattern on front face
      FastLED.clear();
      LED[12].setRGB(3,3,0);
      LED[11].setRGB(0,0,4);
      LED[10].setRGB(0,0,2);
      LED[ 9].setRGB(0,0,1);
      LED_Run = false; LEDshow = true; LED_Task++; break;
    case 1: // rotate LED pattern
      if (LedTick) {
        LedTick = false; LED_Cnt++;
        if (LED_Cnt >= 4) {
          LED[ 7] = LED[11]; LED[ 6] = LED[12]; LED[ 5] = LED[12]; LED[ 4] = LED[11];
          LED[ 0] = LED[10]; LED[ 1] = LED[ 9]; LED[ 2] = LED[ 9]; LED[ 3] = LED[10];
          LEDX[0] = LED[12]; // grab the front right LED rgb values
          LED[12] = LED[11]; LED[11] = LED[10]; LED[10] = LED[ 9]; LED[ 9] = LED[ 8];
          LED[ 8] = LEDX[0];
          LED_Run = false; LEDshow = true; LED_Cnt = 0;
        }
      } break;
  }
}

// --------------------------------------------------------------------------------

void LED_Main() {
  // tasks for manipulating RGB LED patterns
  // Blink clock and modes
  Blink = false; BlinkCnt++; if (BlinkCnt > 30) {BlinkCnt = 0; Blink = true;}
  // switch over-rides
  if ((sw0State == LOW) || (sw1State == LOW)) {
    FastLED.clear();
    if (sw0State == LOW) {LED_SW0();}  // force left side LEDs
    if (sw1State == LOW) {LED_SW1();}  // force right side LEDs
    LED_Run = false; LEDshow = true; LED_Del = 1; return;
  }
  if (LED_Del > 0) {LED_Del--; LED_Run = false; return;} // delay by skipping tasks
  
  if (TEST) {LED_Test(); return;} // TEST mode over-ride

//  Serial.println(LedMode);
  // when performing drive tasks vary rate with LedSpd
  if (LedMode >= 21) {LED_Period = map(LedSpd,PWM_StartMin,255,20,10);}
  
  switch(LedMode) {
    case 0: // doing nothing
      LED_Sleep(); break;
    case 2: // Ready
      LED_Ready(); break;
    case 5: // LTOF range
      LED_Range(); break;
    case 10: // random LED colours
      GetRndCol(2); LED[random(13)].setRGB(Col_R,Col_G,Col_B);
      GetRndCol(4); LED[random(13)].setRGB(Col_R,Col_G,Col_B);
      GetRndCol(8); LED[random(13)].setRGB(Col_R,Col_G,Col_B);
      GetRndCol(2); LED[random(13)].setRGB(Col_R,Col_G,Col_B);
      GetRndCol(4); LED[random(13)].setRGB(Col_R,Col_G,Col_B);
      GetRndCol(8); LED[random(13)].setRGB(Col_R,Col_G,Col_B);
      LED_Run = false; LEDshow = true; LED_Del = 0;
      break;
    case 21: // driving forward
      LED_DriveFwd(); break;
    case 22: // driving sideways to the right
      LED_DriveRht(); break;
    case 23: // driving backward
      LED_DriveBck(); break;
    case 24: // driving sideways to the left
      LED_DriveLft(); break;
    case 25: // driving neutral turn left
      LED_NeutLft(); break;
    case 26: // driving neutral turn right
      LED_NeutRht(); break;
    case 31: // Pixy2 CCC static mode
      LED_Pixy_CCC(); break;
  }
}

// --------------------------------------------------------------------------------

void LED_NeutLft() {
  // generates a neutral turn leftt pattern
  switch (LED_Task) {
    case 0: // create LED pattern on front face
      FastLED.clear();
      LED[ 8].setRGB(3,3,0);
      LED[ 9].setRGB(0,0,4);
      LED[10].setRGB(0,0,2);
      LED[11].setRGB(0,0,1);
      LED_Run = false; LEDshow = true; LED_Task++; break;
    case 1: // rotate LED pattern
      if (LedTick) {
        LedTick = false; LED_Cnt++;
        if (LED_Cnt >= 4) {
          LED[3] = LED[ 8]; LED[2] = LED[ 9]; LED[1] = LED[10]; LED[0] = LED[11];
          LED[7] = LED[ 8]; LED[6] = LED[ 9]; LED[5] = LED[10]; LED[4] = LED[11];
          LEDX[0] = LED[ 8]; // grab the front right LED rgb values
          LED[ 8] = LED[ 9]; LED[ 9] = LED[10]; LED[10] = LED[11]; LED[11] = LED[12];
          LED[12] = LEDX[0];
          LED_Run = false; LEDshow = true; LED_Cnt = 0;
        }
      } break;
  }
}

// --------------------------------------------------------------------------------

void LED_NeutRht() {
  // generates a neutral turn right pattern
  switch (LED_Task) {
    case 0: // create LED pattern
      FastLED.clear();
      LED[12].setRGB(3,3,0);
      LED[11].setRGB(0,0,4);
      LED[10].setRGB(0,0,2);
      LED[ 9].setRGB(0,0,1);
      LED_Run = false; LEDshow = true; LED_Task++; break;
    case 1: // rotate LED pattern
      if (LedTick) {
        LedTick = false; LED_Cnt++;
        if (LED_Cnt >= 4) {
          LED[0] = LED[12]; LED[1] = LED[11]; LED[2] = LED[10]; LED[3] = LED[ 9];
          LED[4] = LED[12]; LED[5] = LED[11]; LED[6] = LED[10]; LED[7] = LED[ 9];
          LEDX[0] = LED[12]; // grab the front left LED rgb values
          LED[12] = LED[11]; LED[11] = LED[10]; LED[10] = LED[ 9]; LED[ 9] = LED[ 8];
          LED[ 8] = LEDX[0];
          LED_Run = false; LEDshow = true; LED_Cnt = 0;
        }
      } break;
  }
}

// --------------------------------------------------------------------------------

void LED_Pixy_CCC() {
  // lights LEDs to match recognised object(s)  
  FastLED.clear();
  if (BlockCnt <= 0) {
    LEDX[0].setRGB(1,1,0);
    switch(LED_Cnt) {
      case 0: LED[10] = LEDX[0]; LED[ 0] = LEDX[0]; LED[ 7] = LEDX[0]; break;
      case 1: LED[ 9] = LEDX[0]; LED[11] = LEDX[0]; LED[ 1] = LEDX[0]; LED[ 6] = LEDX[0]; break;
      case 2: LED[ 8] = LEDX[0]; LED[12] = LEDX[0]; LED[ 2] = LEDX[0]; LED[ 5] = LEDX[0]; break;
      case 3: LED[ 3] = LEDX[0]; LED[ 4] = LEDX[0]; break;
      case 4: LED[ 8] = LEDX[0]; LED[12] = LEDX[0]; LED[ 2] = LEDX[0]; LED[ 5] = LEDX[0]; break;
      case 5: LED[ 9] = LEDX[0]; LED[11] = LEDX[0]; LED[ 1] = LEDX[0]; LED[ 6] = LEDX[0]; break;
    }
    LED_Cnt++; if (LED_Cnt > 5) {LED_Cnt = 0;}
  } else {
    // one or more blocks detected
    int zI;   // colour intensity based on block width
    int zI2;  // zI x 2
    int zI3;  // zI x 3
    int zW;   // temporary width variable
    for (int zb = 0;zb < BlockCnt;zb++) {
      // take each block in turn, map it onto the top LEDs
      // the width or height determines the brightness
      zW = Pixy_mW[zb]; if (zW < Pixy_mH[zb]) {zW = Pixy_mH[zb];}
      zI = map(zW,PixWdMin,315,0,127); zI = constrain(zI,0,127);
      zI2 = zI * 2; zI2 = min(zI2,255);
      zI3 = zI * 3; zI3 = min(zI3,255);
      switch (Pixy_Sg[zb]) {
        case B001: LEDX[0].setRGB( zI3,  0,  0); break; // red
        case B010: LEDX[0].setRGB(   0,zI3,  0); break; // green
        case B011: LEDX[0].setRGB(   0,  0,zI3); break; // blue
        case B100: LEDX[0].setRGB(  zI, zI,  0); break; // yellow
        case B101: LEDX[0].setRGB( zI3, zI,  0); break; // orange
      }
      // is there an assigned colour signature?
      if (SigId == 0) {
        // no assigned colour signature
        switch (zb) {
          case 0: LED[0] = LEDX[0]; LED[7] = LEDX[0]; break;  // colour 1st row
          case 1: LED[1] = LEDX[0]; LED[6] = LEDX[0]; break;  // colour 2nd row
          case 2: LED[2] = LEDX[0]; LED[5] = LEDX[0]; break;  // colour 3ed row
        }
      } else {
        // yes there is an assigned colour signature, does it match
        if (Pixy_Sg[zb] == SigId) {
          // this is the first and therefore largest assigned object
          // we use the size to create a bar graph
          zI = map(zW,PixWdMin,SigHi,0,4); zI = constrain(zI,0,4);
          if (zI > 0) {LED[0] = LEDX[0]; LED[7] = LEDX[0];}
          if (zI > 1) {LED[1] = LEDX[0]; LED[6] = LEDX[0];}
          if (zI > 2) {LED[2] = LEDX[0]; LED[5] = LEDX[0];}
          if (zI > 3) {LED[3] = LEDX[0]; LED[4] = LEDX[0];}
          break;  // exit the ForNext loop
        }
      }
    }
  }
  LED_Run = false; LEDshow = true; LED_Del = 5;
}

// --------------------------------------------------------------------------------

void LED_Range() {
  // drive LEDs in response to LTOF ranging   
  FastLED.clear();
  if (RangeRaw < RangeMax) {
    anyVal = map(Range,RangeMin,RangeMax,110,0);
    if (anyVal >=  10) {LED[ 8].setRGB(1,1,0);}
    if (anyVal >=  20) {LED[ 8].setRGB(5,0,0);}
    if (anyVal >=  30) {LED[ 9].setRGB(1,1,0); LED[ 0].setRGB(1,1,0);  LED[ 7].setRGB(1,1,0);}
    if (anyVal >=  40) {LED[ 9].setRGB(5,0,0); LED[ 0].setRGB(5,0,0);  LED[ 7].setRGB(5,0,0);}
    if (anyVal >=  50) {LED[10].setRGB(1,1,0); LED[ 1].setRGB(1,1,0);  LED[ 6].setRGB(1,1,0);}
    if (anyVal >=  60) {LED[10].setRGB(5,0,0); LED[ 1].setRGB(5,0,0);  LED[ 6].setRGB(5,0,0);}
    if (anyVal >=  70) {LED[11].setRGB(1,1,0); LED[ 2].setRGB(1,1,0);  LED[ 5].setRGB(1,1,0);}
    if (anyVal >=  80) {LED[11].setRGB(5,0,0); LED[ 2].setRGB(5,0,0);  LED[ 5].setRGB(5,0,0);}
    if (anyVal >=  90) {LED[12].setRGB(1,1,0); LED[ 3].setRGB(1,1,0);  LED[ 4].setRGB(1,1,0);}
    if (anyVal >= 100) {LED[12].setRGB(5,0,0); LED[ 3].setRGB(5,0,0);  LED[ 4].setRGB(5,0,0);}
  } else {
    if (Blink) {
      LED[ 8].setRGB(1,0,0); LED[ 9].setRGB(1,0,0); LED[10].setRGB(1,0,0); LED[11].setRGB(1,0,0); LED[12].setRGB(1,0,0);
      LED[ 0].setRGB(1,0,0); LED[ 7].setRGB(1,0,0);
    }
  }
  LED_Run = false; LEDshow = true; LED_Del = 0;
}

// --------------------------------------------------------------------------------

void LED_Ready() {
  // Red walking dots   
  FastLED.clear();
       if (Gear == 1) {LEDX[0].setRGB(0,2,0);}  // green
  else if (Gear == 2) {LEDX[0].setRGB(0,1,1);}  // cyan
  else if (Gear == 3) {LEDX[0].setRGB(0,0,2);}  // blue
  else if (Gear == 4) {LEDX[0].setRGB(1,0,1);}  // purple
  else {LEDX[0].setRGB(4,0,0);}                 // red
  switch(LED_Cnt) {
    case 0: LED[10] = LEDX[0]; break;
    case 1: LED[ 9] = LEDX[0]; LED[11] = LEDX[0]; break;
    case 2: LED[ 8] = LEDX[0]; LED[12] = LEDX[0]; break;
    case 3: LED[ 0] = LEDX[0]; LED[ 7] = LEDX[0]; break;
    case 4: LED[ 1] = LEDX[0]; LED[ 6] = LEDX[0]; break;
    case 5: LED[ 2] = LEDX[0]; LED[ 5] = LEDX[0]; break;
    case 6: LED[ 3] = LEDX[0]; LED[ 4] = LEDX[0]; break;
    default:
      LED_Cnt = -1; break;
  } LED_Cnt++;
  LED_Run = false; LEDshow = true; LED_Del = 3;
}

// --------------------------------------------------------------------------------

void LED_Set_All(byte zR,byte zG,byte zB) {
  // sets all of the LEDs to one colour immediately
  for (byte zL = 0; zL < 13;zL++) {LED[zL].setRGB(zR,zG,zB);}
  FastLED.show();
}

// --------------------------------------------------------------------------------

void LED_Sleep() {
  // Blue slow walking dots up/down
  FastLED.clear(); LEDX[0].setRGB(0,0,1);
  switch(LED_Cnt) {
    case 0: LED[10] = LEDX[0]; LED[ 0] = LEDX[0]; LED[ 7] = LEDX[0]; break;
    case 1: LED[ 9] = LEDX[0]; LED[11] = LEDX[0]; LED[ 1] = LEDX[0]; LED[ 6] = LEDX[0]; break;
    case 2: LED[ 8] = LEDX[0]; LED[12] = LEDX[0]; LED[ 2] = LEDX[0]; LED[ 5] = LEDX[0]; break;
    case 3: LED[ 3] = LEDX[0]; LED[ 4] = LEDX[0]; break;
    case 4: LED[ 8] = LEDX[0]; LED[12] = LEDX[0]; LED[ 2] = LEDX[0]; LED[ 5] = LEDX[0]; break;
    case 5: LED[ 9] = LEDX[0]; LED[11] = LEDX[0]; LED[ 1] = LEDX[0]; LED[ 6] = LEDX[0]; break;
  }
  LED_Cnt++; if (LED_Cnt > 5) {LED_Cnt = 0;}
  LED_Run = false; LEDshow = true; LED_Del = 25;
}

// --------------------------------------------------------------------------------

void LED_Store() {
  // default, slow green walking dots   
  FastLED.clear();
  switch(LED_Cnt) {
    case 0: LED[10].setRGB(0,1,0); break;
    case 1: LED[ 9].setRGB(0,1,0); LED[11].setRGB(0,1,0); break;
    case 2: LED[ 8].setRGB(0,1,0); LED[12].setRGB(0,1,0); break;
    case 3: LED[ 0].setRGB(0,1,0); LED[ 7].setRGB(0,1,0); break;
    case 4: LED[ 1].setRGB(0,1,0); LED[ 6].setRGB(0,1,0); break;
    case 5: LED[ 2].setRGB(0,1,0); LED[ 5].setRGB(0,1,0); break;
    case 6: LED[ 3].setRGB(0,1,0); LED[ 4].setRGB(0,1,0); break;
    default:
      LED_Cnt = -1; break;
  } LED_Cnt++;
  LED_Run = false; LEDshow = true; LED_Del = 12;
}

// --------------------------------------------------------------------------------

void LED_SW0() {
  // called as an over-ride when SW0 is pressed
  LED[0].setRGB(0,0,3); LED[1] = LED[0];  LED[2] = LED[1]; LED[3] = LED[2];
}

// --------------------------------------------------------------------------------

void LED_SW1() {
  // called as an over-ride when SW1 is pressed
  LED[4].setRGB(0,0,3); LED[5] = LED[4];  LED[6] = LED[5]; LED[7] = LED[6];
}

// --------------------------------------------------------------------------------

void LED_Test() {
  // scrolling RGB sequence in TEST mode
  // scroll LEDs up
  LED[12] = LED[11]; LED[11] = LED[10]; LED[10] = LED[9]; LED[9] = LED[8];
  LED[8] = LED[7]; LED[7] = LED[6]; LED[6] = LED[5]; LED[5] = LED[4];
  LED[4] = LED[3]; LED[3] = LED[2]; LED[2] = LED[1]; LED[1] = LED[0];
  
//  LED[0].setRGB(0,0,0); // set LED[0] black
  switch(LED_Task) {
    case 0: LED[0].setRGB(2,0,0); break;
    case 3: LED[0].setRGB(0,2,0); break;
    case 6: LED[0].setRGB(0,0,2); break;
  }
  LED_Task++; if (LED_Task > 8) {LED_Task = 0;}
  LED_Run = false; LEDshow = true; LED_Del = 6;
}

// --------------------------------------------------------------------------------

int16_t map16(int16_t x,int16_t in_min,int16_t in_max,int16_t out_min,int16_t out_max) {
  // standard map function, but for int16_t variables
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

// --------------------------------------------------------------------------------

void MPU_6050_Initialise() {
  // set up gyro hardware
  // By default the MPU-6050 sleeps. So we have to wake it up.
  Wire.beginTransmission(gyro_address);   //Start communication with the address found during search.
  Wire.write(0x6B);         //We want to write to the PWR_MGMT_1 register (6B hex)
  Wire.write(0x00);         //Set the register bits as 00000000 to activate the gyro
  Wire.endTransmission(false);   //End the transmission with the gyro.
  //Set the full scale of the gyro to +/- 250 degrees per second
  Wire.beginTransmission(gyro_address);   //Start communication with the address found during search.
  Wire.write(0x1B);         //We want to write to the GYRO_CONFIG register (1B hex)
  Wire.write(0x00);         //Set the register bits as 00000000 (250dps full scale)
  Wire.endTransmission(false);   //End the transmission with the gyro
  //Set the full scale of the accelerometer to +/- 4g.
  Wire.beginTransmission(gyro_address);   //Start communication with the address found during search.
  Wire.write(0x1C);         //We want to write to the ACCEL_CONFIG register (1A hex)
  Wire.write(0x00);         //Set the register bits as 00000000 (+/- 2g full scale range)
//  Wire.write(0x08);         //Set the register bits as 00001000 (+/- 4g full scale range)
  Wire.endTransmission(false);   //End the transmission with the gyro
//  //Set some filtering to improve the raw data.
//  Wire.beginTransmission(gyro_address);   //Start communication with the address found during search
//  Wire.write(0x1A);         //We want to write to the CONFIG register (1A hex)
//  Wire.write(0x03);         //Set the register bits as 00000011 (Set Digital Low Pass Filter to ~43Hz)
//  Wire.endTransmission(false);   //End the transmission with the gyro 
}

// --------------------------------------------------------------------------------

bool MotorToTgt(int16_t zFL,int16_t zFR,int16_t zRL,int16_t zRR) {
  // moves motor demans towards target values
  // uses intelligent start and stop
  bool zAtTGt = true; int16_t zI = 2;
  if (zFL > PWM_FL) { // increase front left value
         if (zFL > 0) {PWM_FL += zI; if (PWM_FL < PWM_StartMax) {PWM_FL = PWM_StartMax;}}
    else if (zFL < 0) {PWM_FL += zI; if (PWM_FL > zFL) {PWM_FL = zFL;}}
    else {if (PWM_FL < -PWM_StartMax) {PWM_FL += zI;} else {PWM_FL = 0;}}
    zAtTGt = false;
  } else if (zFL < PWM_FL) { // decrease front left value
         if (zFL < 0) {PWM_FL -= zI; if (PWM_FL > -PWM_StartMax) {PWM_FL = -PWM_StartMax;}}
    else if (zFL > 0) {PWM_FL -= zI; if (PWM_FL < zFL) {PWM_FL = zFL;}}
    else {if (PWM_FL > PWM_StartMax) {PWM_FL -= zI;} else {PWM_FL = 0;}}
    zAtTGt = false;
  }
  if (zFR > PWM_FR) { // increase front right value
         if (zFR > 0) {PWM_FR += zI; if (PWM_FR < PWM_StartMax) {PWM_FR = PWM_StartMax;}}
    else if (zFR < 0) {PWM_FR += zI; if (PWM_FR > zFR) {PWM_FR = zFR;}}
    else {if (PWM_FR < -PWM_StartMax) {PWM_FR += zI;} else {PWM_FR = 0;}}
    zAtTGt = false;
  } else if (zFR < PWM_FR) { // decrease front right value
         if (zFR < 0) {PWM_FR -= zI; if (PWM_FR > -PWM_StartMax) {PWM_FR = -PWM_StartMax;}}
    else if (zFR > 0) {PWM_FR -= zI; if (PWM_FR < zFR) {PWM_FR = zFR;}}
    else {if (PWM_FR > PWM_StartMax) {PWM_FR -= zI;} else {PWM_FR = 0;}}
    zAtTGt = false;
  }
  if (zRL > PWM_RL) { // increase rear left value
         if (zRL > 0) {PWM_RL += zI; if (PWM_RL < PWM_StartMax) {PWM_RL = PWM_StartMax;}}
    else if (zRL < 0) {PWM_RL += zI; if (PWM_RL > zRL) {PWM_RL = zRL;}}
    else {if (PWM_RL < -PWM_StartMax) {PWM_RL += zI;} else {PWM_RL = 0;}}
    zAtTGt = false;
  } else if (zRL < PWM_RL) { // decrease rear left value
         if (zRL < 0) {PWM_RL -= zI; if (PWM_RL > -PWM_StartMax) {PWM_RL = -PWM_StartMax;}}
    else if (zRL > 0) {PWM_RL -= zI; if (PWM_RL < zRL) {PWM_RL = zRL;}}
    else {if (PWM_RL > PWM_StartMax) {PWM_RL -= zI;} else {PWM_RL = 0;}}
    zAtTGt = false;
  }
  if (zRR > PWM_RR) { // increase rear right value
         if (zRR > 0) {PWM_RR += zI; if (PWM_RR < PWM_StartMax) {PWM_RR = PWM_StartMax;}}
    else if (zRR < 0) {PWM_RR += zI; if (PWM_RR > zRR) {PWM_RR = zRR;}}
    else {if (PWM_RR < -PWM_StartMax) {PWM_RR += zI;} else {PWM_RR = 0;}}
    zAtTGt = false;
  } else if (zRR < PWM_RR) { // decrease rear right value
         if (zRR < 0) {PWM_RR -= zI; if (PWM_RR > -PWM_StartMax) {PWM_RR = -PWM_StartMax;}}
    else if (zRR > 0) {PWM_RR -= zI; if (PWM_RR < zRR) {PWM_RR = zRR;}}
    else {if (PWM_RR > PWM_StartMax) {PWM_RR -= zI;} else {PWM_RR = 0;}}
    zAtTGt = false;
  }
  
//  Serial.println(String(PWM_FL) + "," + String(PWM_FR) + "," + String(PWM_RL) + "," + String(PWM_RR));
  return zAtTGt;  // zAtTGt == true if at target
}

// --------------------------------------------------------------------------------

void MotorPWM (int16_t zFL,int16_t zFR,int16_t zRL,int16_t zRR) {
  // sets the four H-bridge PWM values for drive control
  // zV > 0 - move wheel in direction for forward motion
  // zV = 0 - lock wheel in brake condition
  // zV < 0 - move wheel in direction for reverse motion

  // store variables for display and counters
  PWM_FL = zFL; PWM_FR = zFR; PWM_RL = zRL; PWM_RR = zRR;
  if (!PwmEn) {zFL = 0; zFR = 0; zRL = 0; zRR = 0;} // motor PWM is disabled
}

// --------------------------------------------------------------------------------

void MPU_print_Values() {
  // send values to the serial monitor
  Serial.print("Xacc "+ String(accelerometer_data_rawX_i));
  Serial.print("\tYacc "+ String(accelerometer_data_rawY_i));
  Serial.println("\tZgyro "+ String(gyro_Z_data_raw));
}

// --------------------------------------------------------------------------------

void MPU_read_AccelAll() {
  // read the X,Y,Z-axis acceleration values from the MPU
  Wire.beginTransmission(gyro_address);
  Wire.write(0x3B);  // starting with register 0x3B (ACCEL_XOUT_H)
//  Wire.endTransmission(false);
  I2C_Error = Wire.endTransmission(false);
  if (I2C_Error > 0) {return;}
//  delayMicroseconds(20);
//  Wire.requestFrom(gyro_address,4,true);  // request 1 register read
  Wire.requestFrom(gyro_address,4);  // request 1 register read
//  if (Wire.available() != 4) {return;}    // ensure that the correct number of bytes have been received
  while(Wire.available() < 4);  // wait for values to be read
  accelerometer_data_rawX_i = Wire.read()<<8|Wire.read();  // 0x3B (ACCEL_XOUT_H) & 0x3C (ACCEL_XOUT_L)
  accelerometer_data_rawY_i = Wire.read()<<8|Wire.read();  // 0x3D (ACCEL_YOUT_H) & 0x3E (ACCEL_YOUT_L)
//  accelerometer_data_rawZ_i = Wire.read()<<8|Wire.read();  // 0x3F (ACCEL_ZOUT_H) & 0x40 (ACCEL_ZOUT_L)
  // record values
  accelerometer_data_rawX_OALast = accelerometer_data_rawX_OA;
  accelerometer_data_rawY_OALast = accelerometer_data_rawY_OA;
  // correct with calibrated offset values
  accelerometer_data_rawX_OA = accelerometer_data_rawX_i + acc_cal_valueX; //Add the accelerometer calibration value
  accelerometer_data_rawY_OA = accelerometer_data_rawY_i + acc_cal_valueY; //Add the accelerometer calibration value
}

// --------------------------------------------------------------------------------

void MPU_read_Rotation() {
  // read the raw pitch and yaw gyro values  
  Wire.beginTransmission(gyro_address);   //Start communication with the gyro
  Wire.write(0x47);                       //Start reading the register GYRO_ZOUT_H
  I2C_Error = Wire.endTransmission(false);
  if (I2C_Error > 0) {return;}

//  delayMicroseconds(20);
//  Wire.requestFrom(gyro_address, 2,true);         // Request 2 bytes from the gyro
  Wire.requestFrom(gyro_address, 2);         // Request 2 bytes from the gyro
//  if (Wire.available() != 2) {return;} // ensure that the correct number of bytes have been received
  while(Wire.available() < 2);  // wait for values to be read
  gyro_Z_data_raw = Wire.read()<<8|Wire.read();   //Combine the two bytes to make one integer
}

// ---------------------------------------------------------------------------------

void PwmAttach(uint8_t zCh) {
  // attached the micros internal PWM counter to an output pin
  switch (zCh) {
    case 0: ledcAttachPin(PinBckLftA,PwmChBckLftA); break;
    case 1: ledcAttachPin(PinBckLftB,PwmChBckLftB); break;
    case 2: ledcAttachPin(PinBckRhtA,PwmChBckRhtA); break;
    case 3: ledcAttachPin(PinBckRhtB,PwmChBckRhtB); break;
    case 4: ledcAttachPin(PinFntLftA,PwmChFntLftA); break;
    case 5: ledcAttachPin(PinFntLftB,PwmChFntLftB); break;
    case 6: ledcAttachPin(PinFntRhtA,PwmChFntRhtA); break;
    case 7: ledcAttachPin(PinFntRhtB,PwmChFntRhtB); break;
  }
}

// ---------------------------------------------------------------------------------

void PwmSetup(uint8_t zCh) {
  // setup the PWM counter to an output pin
  switch (zCh) {
    case 0: ledcSetup(PwmChBckLftA, PwmFreq, PwmBits); break;
    case 1: ledcSetup(PwmChBckLftB, PwmFreq, PwmBits); break;
    case 2: ledcSetup(PwmChBckRhtA, PwmFreq, PwmBits); break;
    case 3: ledcSetup(PwmChBckRhtB, PwmFreq, PwmBits); break;
    case 4: ledcSetup(PwmChFntLftA, PwmFreq, PwmBits); break;
    case 5: ledcSetup(PwmChFntLftB, PwmFreq, PwmBits); break;
    case 6: ledcSetup(PwmChFntRhtA, PwmFreq, PwmBits); break;
    case 7: ledcSetup(PwmChFntRhtB, PwmFreq, PwmBits); break;
  }
}

// --------------------------------------------------------------------------------

void readSerial() {
  // reads characters from the serial port and responds to commands
  // set PrintTgt to indicate source as serial port for responses
  keyVal = Serial.read();
  if (keyVal != -1) {SerialRx = true; PrintTgt = 0; decodeKey(keyVal);}
}

// ---------------------------------------------------------------------------------

void Read_Slots() {
  // called at the beginning of each main loop
  // assynchronously tracks slot pulse triggers
  // FL sensor:
  if (digitalRead(INT_P36) != IntP36State) {
    IntMicros[0] = micros(); IntP36State = !IntP36State; IntP36Cnt+= DIR_FL;
    IntPeriods[0] = (IntPeriods[0] + IntMicros[0] - IntLast[0])/2; IntLast[0]= IntMicros[0];
    // record period for 1/4 revolution (5 slots, 10 edges)
    if (IntLast40[0] == 0) {
      // 1st pulse after reset
      IntLast40[0] = millis(); IntPeriods[0] = 0;
    } else {Int36Rev++; if (Int36Rev >= 10) {
        // we determine rpm over a 1/4 cycle, ie. 10 edges in a 20 slot wheel
        // int rpm is 10 times larger than actual for display purposes
        if (millis() != IntLast40[0]) { // avoid divide by zero error!
          Int36RPM = (DIR_FL * 150000)/int16_t(millis() - IntLast40[0]);
          IntLast40[0] = millis(); Int36Rev = 0;
//      Serial.println(0.1 * Int36RPM);
        }
      }
    } IntP36Ch = true;  // flag change has occurred, wheel is moving
  }
  
  // FR sensor:
  if (digitalRead(INT_P34) != IntP34State) {
    IntMicros[1] = micros(); IntP34State = !IntP34State; IntP34Cnt+= DIR_FR;
    IntPeriods[1] = (IntPeriods[1] + IntMicros[1] - IntLast[1])/2; IntLast[1]= IntMicros[1];
    // record period for 1/4 revolution (5 slots, 10 edges)
    if (IntLast40[1] == 0) {
      // 1st pulse after reset
      IntLast40[1] = millis(); IntPeriods[1] = 0;
    } else {Int34Rev++; if (Int34Rev >= 10) {
        // we determine rpm over a 1/4 cycle, ie. 10 edges in a 20 slot wheel
        // int rpm is 10 times larger than actual for display purposes
        if (millis() != IntLast40[1]) { // avoid divide by zero error!
          Int34RPM = (DIR_FR * 150000)/int16_t(millis() - IntLast40[1]);
          IntLast40[1] = millis(); Int34Rev = 0;
        }
      }
    }IntP34Ch = true;  // flag change has occurred, wheel is moving
  }
  
  // RL sensor:
  if (digitalRead(INT_P35) != IntP35State) {
    IntMicros[2] = micros(); IntP35State = !IntP35State; IntP35Cnt+= DIR_RL;
    IntPeriods[2] = (IntPeriods[2] + IntMicros[2] - IntLast[2])/2; IntLast[2]= IntMicros[2];
    // record period for 1/4 revolution (5 slots, 10 edges)
    if (IntLast40[2] == 0) {
      // 1st pulse after reset
      IntLast40[2] = millis(); IntPeriods[2] = 0;
    } else {Int35Rev++; if (Int35Rev >= 10) {
        // we determine rpm over a 1/4 cycle, ie. 10 edges in a 20 slot wheel
        // int rpm is 10 times larger than actual for display purposes
        if (millis() != IntLast40[2]) { // avoid divide by zero error!
          Int35RPM = (DIR_RL * 150000)/int16_t(millis() - IntLast40[2]);
          IntLast40[2] = millis(); Int35Rev = 0;
        }
      }
    } IntP35Ch = true;  // flag change has occurred, wheel is moving
  }
  
  // RR sensor:
  if (digitalRead(INT_P39) != IntP39State) {
    IntMicros[3] = micros(); IntP39State = !IntP39State; IntP39Cnt+= DIR_RR;
    IntPeriods[3] = (IntPeriods[3] + IntMicros[3] - IntLast[3])/2; IntLast[3]= IntMicros[3];
    // record period for 1/4 revolution (5 slots, 10 edges)
    if (IntLast40[3] == 0) {
      // 1st pulse after reset
      IntLast40[3] = millis(); IntPeriods[3] = 0;
    } else {Int39Rev++; if (Int39Rev >= 10) {
        // we determine rpm over a 1/4 cycle, ie. 10 edges in a 20 slot wheel
        // int rpm is 10 times larger than actual for display purposes
        if (millis() != IntLast40[3]) { // avoid divide by zero error!
          Int39RPM = (DIR_RR * 150000)/int16_t(millis() - IntLast40[3]);
          IntLast40[3] = millis(); Int39Rev = 0;
        }
      }
    } IntP39Ch = true;  // flag change has occurred, wheel is moving
  }
}

// ---------------------------------------------------------------------------------

void readSW0() {
  // read the left button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the current task
  // SW0 is read every 16.666ms (60Hz) to debounce
  sw0State = digitalRead(sw0Pin); // record button state
  if (sw0_Nop) {return;}          // block this function whilst == true
  
  bool sw0Clr = false;
  if (sw0State == LOW) {
    //##############################################################################
    //
    // SW0 button is pressed down
    //
    //##############################################################################
    if (sw0Wup) {return;}
    
    if (sw0LastState == HIGH) {
      // SW0 has just been pressed down
//      Serial.println("SW0 Hi-LO");
      if (TEST) {Serial.println("SW0 Hi-LO");}
      sw0DwnTime = 0; // restart button down time counter
      sw0Cnt++;       // count on the falling edge
      sw0Timer = 0;   // reset the timer
//      if (WiFiEn) {SetMainMode(0); return;}
      if (MainMode == 4) {
        switch (MainTask) {
          case 40:  // toggle Settings selection options due to < NO >
            SetSel++; if (SetSel > 1) {SetSel = 0;}
            sw0Cnt = 0; DispDel = 0; DispCnt = 0; break;
          case 41:  // camera lux adjustment, reduce its value
            if (DispVal > 20) {DispVal-= 5;}
            sw0Cnt = 0; DispDel = 0; DispCnt = 0; break;
        }
      }
      else if (MainMode == 99) {
        // toggle MainMode selection options due to < NO >
        ModeSel++; if (ModeSel > 4) {ModeSel = 0;}
        sw0Cnt = 0; DispDel = 0; DispCnt = 0;
      }
    } else {
      // whilst the button is down adjust the timer at 20ms steps
      // if Wii is not connected test for a long press to invoke demo mode
      sw0DwnTime++; // track button pressed time
      if (TEST) {
        // look for long press > 1 sec in TEST mode
        if (sw0Timer == 120) {
          // 2 second timeout
          sw0Cnt = 0; sw0Timer = 0;
        }
      } else {
        if (sw0Timer > 30) {
          // button SW0 held down for > 0.5 second
          SetMainMode(99);  // switch to mode select
          // block further actions until released
          sw0Cnt = 0; sw0Timer = 0; sw0Wup = true;
        }
      }
    }
  } else {
    //##############################################################################
    //
    // SW0 button is released
    //
    //##############################################################################
    if (sw0Wup) {
      // waited for button release
      sw0Wup = false; sw0Cnt = 0; sw0Timer = 0;
    } else {
      if (sw0LastState == LOW) {
        // SW0 has just been released
//        Serial.println("SW0 Lo-Hi");
        if (TEST) {Serial.println("SW0 Lo-Hi");}

        // if not asleep we change the display mode
        if (MainMode != 0) {
          // SW0 display functions are dependant on MainMode
          // if MainMode == 0 do nothing!
          if (MainMode == 1) {
            // Drive mode, cycle display options
            switch (DispMode) {
              case  1: DispMode = 10; break;  // Drive Mode
              case 10: DispMode = 11; break;  // Motor PWM
              case 11: DispMode = 12; break;  // Slot Periods
              case 12: DispMode = 13; break;  // Slot Counts
              case 13: DispMode = 14; break;  // Slot RPM
              case 14: DispMode =  1; break;  // Tasks
             }
          }
          else if (MainMode == 2) {
            // LTOF mode, SW0 cycles display options
            switch (DispMode) {
              case 20: DispMode = 21; break;  // LTOF ranging
              case 21: DispMode = 11; break;  // Motor PWM
              case 11: DispMode = 13; break;  // Slot counts
              case 13: DispMode = 20; break;  // LTOF Mode
              default: DispMode = 20; break;  // LTOF Mode
            }
          }
          else if (MainMode == 3) {
            // Pixy2 mode, cycle display options
             switch (DispMode) {
              case 30: DispMode = 31; break;  // Pixy2 cam data
              case 31: DispMode = 32; break;  // Pixy2 blocks
              case 32: DispMode = 11; break;  // Motor PWM
              case 11: DispMode = 13; break;  // Slot counts
              default: DispMode = 30; break;  // Pixy2 Mode
            }
          }
          else if (MainMode == 4) {
            // Settings mode, action depends on task
            switch (MainTask) {
              case 41: break; // set camera lux down
              case 42: break; // set LED Lights OFF
            }
          }
          DispDel = 0; DispCnt = 0; DispNxtCnt = 0;
        }
      }
      if (sw0Timer > 60) {
        // button released for 1 sec so assume valid button counts
      //    Serial.print(F("swCnt = ")); Serial.println(swCnt);
        if (TEST && (sw0Cnt > 0)) {
          // single button press during test mode
        }
        sw0Cnt = 0; sw0Timer = 0;
      }
      sw0DwnTime = 0;
    }
  }
  sw0LastState = sw0State; // record current state
  if (sw0Cnt > 0) {sw0Timer++;}
}

// ---------------------------------------------------------------------------------

void readSW1() {
  // read the right button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the WiFi mode
  // SW1 is read every 16.666ms (60Hz) to debounce
  sw1State = digitalRead(sw1Pin); // record button state
  if (sw1_Nop) {return;}          // block this function whilst == true
  
  if (sw1State == LOW) {
    //##############################################################################
    //
    // SW1 button is pressed down
    //
    //##############################################################################
    if (sw1Wup) {return;}
    
    if (sw1LastState == HIGH) {
      // SW1 has just been pressed down
//      Serial.println("SW1 Hi-LO");
      if (TEST) {Serial.println("SW1 Hi-LO");}
      sw1DwnTime = 0; // restart button down time counter
      sw1Cnt++; // count on the falling edge
      sw1Timer = 0; // reset the timer
      if (WiFiEn) {sw1Cnt = 0; SetMainMode(0); return;}
      if (TEST) {
        // put TEST code here
      } else {
        // not in TEST mode
        if (MainMode == 4) {
          // if in MainMode 4 then YES is selected
          switch (MainTask) {
            case 40: // YES is selected
              SetMainTask(41 + SetSel); // enter a setting adjustment tasks
              sw1Wup = true; break;
            case 41: // increase camera lux 
              if (DispVal < 200) {DispVal+= 5;}
              sw1Cnt = 0; DispDel = 0; DispCnt = 0; break;
            case 42: break; // turn LED lights ON
          }
          sw1Cnt = 0; sw1Timer = 0;
        } else if (MainMode == 99) {
          // if in MainMode 99 then YES is selected
          SetMainMode(ModeSel); sw1Wup = true;
          sw1Cnt = 0; sw1Timer = 0;
        } else if (MainMode > 0) {
          // not in MainMode == 99 or sleep so show count on display
          if (MainMode == 1) {
            // Drive mode, counting button clicks
            if (sw1Cnt > 1) {sw1Cnt = 0;}
            switch (sw1Cnt) {
              case 0: DisplayText2S24(String(sw1Cnt),"Inactive"); break;
              case 1: DisplayText2S24(String(sw1Cnt),"Demo"); break;
            }
          }
          else if (MainMode == 2) {
            // LTOF mode, counting button clicks
            if (sw1Cnt > 4) {sw1Cnt = 1;}
            switch (sw1Cnt) {
              case 1: DisplayText2S24(String(sw1Cnt),"LTOF Range"); break;
              case 2: DisplayText2S24(String(sw1Cnt),"Back Away"); break;
              case 3: DisplayText2S24(String(sw1Cnt),"Track"); break;
              case 4: DisplayText2S24(String(sw1Cnt),"Autonomous"); break;
            }
          }
          else if (MainMode == 3) {
            // Pixy2 mode, counting button clicks
            if (sw1Cnt > 6) {sw1Cnt = 1;}
            switch (sw1Cnt) {
              case 1: DisplayText2S24(String(sw1Cnt),"Pixy  Static"); break;
              case 2: DisplayText2S24(String(sw1Cnt),"Pan & Tilt"); break;
              case 3: DisplayText2S24(String(sw1Cnt),"Track / Turn"); break;
              case 4: DisplayText2S24(String(sw1Cnt),"Pixy  Play"); break;
              case 5: DisplayText2S24(String(sw1Cnt),"Pixy  Fast"); break;
              case 6: DisplayText2S24(String(sw1Cnt),"Pixy  Goalie"); break;
            }
          }
        }
      }
    } else {
      sw1DwnTime++; // track button pressed time
      if(sw1Timer >= 60) {
        // button held down for >= 1 second
        if (MainMode == 3) {
          // in Pixy2 mode so toggle lights ON/OFF
          if (Lights) {
            Lights = false; Lamps(0); DisplayText2S24("Lights","OFF");
          } else {
            Lights = true; Lamps(1); DisplayText2S24("Lights","ON");
          }
        }
        sw1Cnt = 0; sw1Timer = 0;
      }
    }
  } else {
    //##############################################################################
    //
    // sw1State == HIGH as SW1 button is not pressed
    //
    //##############################################################################
    if (sw1Wup) {
      // release locked-out state
      sw1Wup = false;
      sw1Cnt = 0; sw1Timer = 0;
    } else {
      // normal SW1 release
      if (sw1LastState == LOW) {
        // SW1 has just been released
//        Serial.println("SW1 Lo-Hi");
        if (TEST) {Serial.println("SW1 Lo-Hi");}
      }
      if (!TEST) {
        // not in TEST mode so set task based on count value
        if ((sw1Timer > 60) && (sw1Cnt > 0)) {
          // 1 sec timeout + count > 0
          // task select will depend on MainMode
          switch (MainMode) {
            case 1: // Drive mode, tasks 10 - 11
              sw1Cnt = constrain(sw1Cnt,0,1); SetMainTask(10 + sw1Cnt);
              break;
            case 2: // LTOF mode, tasks 21 - 24
              sw1Cnt = constrain(sw1Cnt,1,4); SetMainTask(20 + sw1Cnt);
              break;
            case 3: // Pixy2 mode, tasks 31 - 36
              bool zL = Lights; Lights = false; // mask lights status when changing Pixy2 modes
              sw1Cnt = constrain(sw1Cnt,1,6); SetMainTask(30 + sw1Cnt);
              Lights = zL; break; // reinstate Lights status
          } 
          DispDel = 0; DispCnt = 0; // update display immediately
          sw1Cnt = 0; sw1Timer = 0; // clear events
        }
      }
    }
  }
  sw1LastState = sw1State; // record current state
  if (sw1Cnt > 0) {sw1Timer++;}
}

// ---------------------------------------------------------------------------------

void readWiFiRx() {
  // called from the main loop every cycle to check for ESP-NOW data  
  // a state machine is used to separate the data blocks by header type
//  DispDel = 500;  // hold off display updates for 10 sec
  switch(Rx_Task) {
    case 0: // determine the type of data to handle
           if (Rx_Buff.ESPdata[Rx_Pnt] == 'B') {Rx_Task = 3; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'C') {WiiType = 'C'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'N') {WiiType = 'N'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '-') {WiiType = '-'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '$') {Rx_Task = 2; break;}
      // oh dear! bloick header not recognised so discard the block and flag error
      else {WiFiRxErr++; WiFiRx = false;}
      break;
      
    case 1: // load Wii I2C data
      // after the header byte there are 6 data bytes + checksum
      // perform XOR on 6 bytes to check for errors
      if (Rx_Buff.ESPdata[7] == (Rx_Buff.ESPdata[1] ^ Rx_Buff.ESPdata[2] ^ Rx_Buff.ESPdata[3] ^ Rx_Buff.ESPdata[4] ^ Rx_Buff.ESPdata[5] ^ Rx_Buff.ESPdata[6])) {
        // data has passed the checksum test
        RxWiFi[0] = Rx_Buff.ESPdata[1];
        RxWiFi[1] = Rx_Buff.ESPdata[2];
        RxWiFi[2] = Rx_Buff.ESPdata[3];
        RxWiFi[3] = Rx_Buff.ESPdata[4];
        RxWiFi[4] = Rx_Buff.ESPdata[5];
        RxWiFi[5] = Rx_Buff.ESPdata[6];

        // if Nunchuk data strip off the JX,JY and CZ values
        RxRec = true; // indicate a valid frame
        if (WiiType == 'N') {
          JoyX = RxWiFi[0];   // joystick X 0-255
          JoyY = RxWiFi[1];   // joystick Y 0-255
          CZ = RxWiFi[5] & 3; // CZ buttons, C = 2/0, Z = 1/0, active low
        } else {
          // not Nunchuk so set defaults
          JoyX = 128; JoyY = 128; CZ = 3;
        }
      } else {
        WiFiRxErr++;
      }
      Rx_Task = 0;  // reset the task pointer to lok at another block if needed
      Rx_Pnt += 8;  // increase the data pointer beyond this block
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0;} // reached end of buffer
      break;

    case 2: // load text data as if from a serial port
      // read characters up to the end of the buffer
      Rx_Pnt++;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0; break;} // reached end of buffer
      keyVal = Rx_Buff.ESPdata[Rx_Pnt]; PrintTgt = 1; decodeKey(keyVal);
//      Serial.write(keyVal);
      break;

    case 3: // load binary data block
      // data length is specified in second byte
      Rx_Pnt++;
      // at the moment just jump over it, binary transfer is not needed
      Rx_Pnt += Rx_Buff.ESPdata[Rx_Pnt] + 1;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; break;} // reached end of buffer
      break;
  }
}

// ---------------------------------------------------------------------------------

void readWii() {
  // called every 20ms, but wireless transceiver only transmits every 40ms
  // considers data received from the Wii Nunchuk controller over WiFi and 
  // manages all movements demanded by the controller
  if (readWiiCall) {return;}  // avoid stack overflow
  
  readWiiCall = true; // prevent re-entry of this function
  
  if (RxRec) {
  //##############################################################################
  //
  //  WiFi Nunchuk data received
  //
  //##############################################################################
    // Wii data received but we may not have enabled WiFi mode yet
    RxRec = false; RTxTimeout = 25; // set timeout to approx 500ms
//    PrintTx += "JX" + String(JoyX) + " JY" + String(JoyY) + " CZ" + String(CZ) + "\n";
//    Serial.print("CZ "); Serial.println(CZ);
    if (!WiFiEn) {
  //##############################################################################
  //
  //  WiFi is currently disabled, listening
  //
  //##############################################################################
      // we have not received enough 'C' buttons to make the system active
      // the 'C' button alone must be held in to activate the Wii WiFi link
      if (CZ == 1) {
        // C button only is being pressed
        if (WiFiCntC < 1) {DisplayText2("Waking","Up!");}
        WiFiCntC += 2;
//        Serial.println("WiFiCnt = " + String(WiFiCntC));
        if (WiFiCntC > 60) {
          // going active, user has held in 'C' button for sufficient time
//          Serial.println("WiFiEn = true");
          SetMainMode(1); PwmEn = true; // go to Drive mode
          PixyBotActive = true; WiFiEn = true; C_Dn = 8;
//          if (MainTask != 0) {SetMainTask(0);}
//          if (!AtRest) {SetMainTask(98); LEDMthTask = -1;} // move to squat (default)
          // set defaults here
//          GoToStand(50);  // go to standing position
          synchLoopTimers(0); autoOFF = 10;
//          Gear = 1;
        } else {
          // still counting C buttons
          LEDVal = WiFiCntC/10; SetLedMode(10);
        }
      } else {
        // C button not pressed or released early so check for a inadequate count
        if (WiFiCntC > 0) {
          // reduce the C-button counter, which has to be sustained to enable WiFi mode
          WiFiCntC--; LEDVal = 1 + (WiFiCntC/10);
          if (WiFiCntC < 1) {
            SetLedMode(0); DisplayText2("Wakeup","Cancelled");
            SetMainMode(0);
          } else {SetLedMode(10);}
        }
      }
    } else {
      
  //##############################################################################
  //
  //  WiFi is enabled
  //
  //##############################################################################
      // WiFi enabled so check for power down
      // CZ button stats are active LOW
      if (CZ == 0) {
        // both C and Z buttons are being pressed so move towards powering down
        if (WiFiCntC > 0) {
          if (WiFiCntC == 64) {DisplayText2("Going To","Sleep..."); DispMode = 0;}
          WiFiCntC -= 2;
  //        Serial.println("WiFiCnt = " + String(WiFiCntC));
          if (WiFiCntC <= 0) {
            // power-down condition satisfied, so go to rest
//            Serial.println("WiFi mode disabled!");
            WiFiCntC = 0; WiFiDisable();
            GoToSleep(); // go to the resting state
            // set defaults here for sleep mode
             PixyBotActive = false; SetLedMode(0);
             SetMainMode(0);
          } else {
            // still counting down CZ buttons
            SetLedMode(10);
          }
        }
      } else {
        // Check CZ buttons to see if a speed change is wanted
        if ((CZ & 2) == 0) {
          // C button is pressed, assume a speed change is needed
          C_Dn++;
          if (C_Dn == 2) {
            C_Dn = 8;
            if (Gear < 5) {Gear++; SetMaxSpeed();}
            if (DispMode != 15) {DisplayText2S24(String(Gear),"Gear");}
          }
//          Serial.println("Gear =" + String(Gear));
        } else {C_Dn = 0;} // C button released
        // Check Z button. It has two functions. Short press change speed, but if
        // held in it changes the mode of the joysticks
        if ((CZ & 1) == 0) {
          // Z button is pressed
          Z_Dn++; if (Z_Dn >= 12) {
            if (Z_Dn == 12) {
              Z_Mode = !Z_Mode; // toggle joystick motion mode flag
              Border = true; // turn on boarder
              if (Z_Mode) {DisplayText2("Slide","Drive");}
              else {DisplayText2("Normal","Drive");} DispDel = 12;        // short display 480ms
              Border = false; // thick boarder
            } Z_Dn = 50;        // Z_Dn capped at 50, held for >= 2 seconds
          }
//          Serial.print("Z_Dn "); Serial.println(Z_Dn);
        } else {
          // Z button is not pressed
          if ((Z_Dn > 1) && (Z_Dn <= 10)) {
            // held down for less than 400 ms @ 40ms Rx cycle
            if (Gear > 1) {Gear--; SetMaxSpeed();}
            if (DispMode != 15) {DisplayText2S24(String(Gear),"Gear");}
          } Z_Dn = 0; // Z button released so clear down counter
        }
//        Serial.print("Z_Dn "); Serial.println(Z_Dn);

        if (WiFiCntC < 60) {
          // C-button counter is not at full strength, assume WiFi off released prematurely
          WiFiCntC++; LEDVal = 1 + (WiFiCntC/10);
          if (WiFiCntC == 60) {LED_Task = -1;} else {LED_Task = -10;}
        }
      }
    }
      
  //##############################################################################
  //
  //  Wii joystick demands
  //
  //##############################################################################
    // joystick movements are ignored if:
    // PixyBot is not active
    // Wii Joystick has not sent new data
    // a MainTask is running
    if (PixyBotActive && WiFiEn && JoyActive && (MainTask == 0)) {
      // read the X,Y joysticks and drive if demanded
      JoyActive = false; // prevent stack overflow during moves that call loop()
      // respond to joystick demands every 40ms (25Hz)
      // note direction of travel will only change once stick has been centred
      if ((JoyX >= DbLLX) && (JoyX <= DbULX) && (JoyY >= DbLLY) && (JoyY <= DbULY)) {
        // joystick is in centre deadband position
        JoyMode = 0; // reset the direction of travel mode
//          Serial.print(F("#"));
        if (Drive != 0) {
//          Serial.println(F("Stopping"));
          DithON = false; SetLedMode(2);
          // stop driving if active, but remember direction
          DriveLast = Drive; Drive = 0;
//          Serial.println("MotorPWM (0,0,0,0)");
          MotorPWM (0,0,0,0); Steer = 0;
        }
      } else {
        // not in centre deadband position
        if (JoyMode == 0) {
          // in normal mode a joystick selection locks us into that mode, until released
          autoRest = 0; // reset user time-out
          // we apply a deadband to the joystick before triggering movement
          // there are potentially 8 directions of travel
          if (!Z_Mode) {
            // Normal drive mode
                 if (JoyY > DbULY) {JoyMode = 1;} // drive forward
            else if (JoyY < DbLLY) {JoyMode = 5;} // drive backwards
            else if (JoyX > DbULX) {JoyMode = 3;} // neutral turn right
            else if (JoyX < DbLLX) {JoyMode = 7;} // neutral turn left
          }
        }
        if (Z_Mode) {
          // Slide drive mode, joystick can change direction of travel at any time
          // JoyModes range from 10 to 18 in this active Z_Mode
          // priority testing goes to forward/reverse and left/right directions
          if (JoyY > DbULY) {
                 // joystick pushed forward
                 if (JoyX < DbLLX) {JoyMode = 18;}  // slide to the NW
            else if (JoyX > DbULX) {JoyMode = 12;}  // slide to the NE
            else {JoyMode = 11;}                    // drive forward
          }
          else if (JoyY < DbLLY) {
                 // joystick pulled backwards
                 if (JoyX < DbLLX) {JoyMode = 16;}  // slide to the SW
            else if (JoyX > DbULX) {JoyMode = 14;}  // slide to the SE
            else {JoyMode = 15;}                    // drive backwards
          }
          else {
                 if (JoyX > DbULX) {JoyMode = 13;}       // slide right
            else if (JoyX < DbLLX) {JoyMode = 17;}       // slide left
          }
        }
        autoRest = 0;     // reset user time-out
        switch (JoyMode) {
          case 1: JoyMoveFwd(); break;
          case 3: JoyTurnRgt(); break;
          case 5: JoyMoveBwd(); break;
          case 7: JoyTurnLft(); break;

          case 11: JoyMoveFwd(); break;
          case 12: JoySlideNE(); break;
          case 13: JoySlideRgt(); break;
          case 14: JoySlideSE(); break;
          case 15: JoyMoveBwd(); break;
          case 16: JoySlideSW(); break;
          case 17: JoySlideLft(); break;
          case 18: JoySlideNW(); break;
        }
      } JoyActive = true; // re-enable this code block
    }
  } else {
  //##############################################################################
  //
  //  No WiFi Nunchuk received recently
  //
  //##############################################################################
    if (WiFiEn) {
      // HexBot is enabled for Wii WiFi control
      // maintain Wii enabled counter
      if ((WiFiCntC < 64) && (CZ != 0)) {
        WiFiCntC++; LEDVal = WiFiCntC/10;
        if (WiFiCntC == 64) {LED_Task = -1;} else {LED_Task = -10;}
      }
    }
  }

  if (WiFiEn) {
  //##############################################################################
  //
  //  WiFi background tasks when WiFi is enabled
  //
  //##############################################################################
    // only do these functions if WiFi is enabled
    if (MoveState == 20) {
      // long un-demanded moves like bowing are watched until they complete
      MoveState = 0; LED_Task = -1;
    }
    
    if (MoveState == -1) {
      // at the end of a joystick move we want to set things back to normal
      LED_Task = -1;
//      SetSpeedDefaults();
      MoveState = 0;
    }
  }
  
  readWiiCall = false; // re-allow calls to this function
//  Serial.print("CntZ "); Serial.println(WiFiCntZ);
//   Serial.println(MoveState);
}

// ---------------------------------------------------------------------------------

void Reset_Slots() {
  // called before using the slot sensors to count pulses
  // FL sensor:
  IntP36State = digitalRead(INT_P36); IntP36Ch = false; IntPeriods[0] = 0;
  IntMicros[0] = micros(); IntP36Cnt = 0; IntLast40[0] = 0; Int36Rev = 0; Int36RPM = 0;
 // FR sensor:
  IntP34State = digitalRead(INT_P34); IntP34Ch = false; IntPeriods[1] = 0;
  IntMicros[1] = micros(); IntP34Cnt = 0; IntLast40[1] = 0; Int34Rev = 0; Int34RPM = 0;
  // RL sensor:
  IntP35State = digitalRead(INT_P35); IntP35Ch = false; IntPeriods[2] = 0;
  IntMicros[2] = micros(); IntP35Cnt = 0; IntLast40[2] = 0; Int35Rev = 0; Int35RPM = 0;
  // RR sensor:
  IntP39State = digitalRead(INT_P39); IntP39Ch = false; IntPeriods[3] = 0;
  IntMicros[3] = micros(); IntP39Cnt = 0; IntLast40[3] = 0; Int39Rev = 0; Int39RPM = 0;
}

// --------------------------------------------------------------------------------

void RxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Rx buffer of length zlen
  Rx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    Rx_Chk = Rx_Chk ^ Rx_Buff.ESPdata[zP];
  }
}

// --------------------------------------------------------------------------------

void SetLedMode(int16_t zM) {
  // sets the LED mode and clears the counter
  // if zM matches current LedMode then do nothing
  // if zM is -ve then reset it
  if (zM == LedMode) {return;}
  
  LedMode = abs(zM); LED_Cnt = 0; LED_Task = 0; LED_Period = 20;
//  Serial.println("SetLedMode" + String(LedMode));
}

// --------------------------------------------------------------------------------

void SetMainMode(uint8_t zM) {
  // sets the default conditions for a new mode
  // MainMode affects display functions and MainTask settings
  //  0 = Sleep mode, default
  //  1 = Drive mode
  //  2 = LTOF mode
  //  3 = Pixy2 mode
  // 99 = mode select
  // now set mode specific items
  ModeSel = MainMode; MainMode = zM;
  if (MainMode > 0) {Display_setBrightness(Brightness);}
  switch (zM) {
    case 0: // sleep mode
      DispMode = 0; PixyBotActive = false; WiFiEn = false;
      SetMainTask(0); SetLedMode(0); break;
    case 1: // Drive mode & WiFi Nunchuk mode
      DispMode = 10; PixyBotActive = true;
      SetMainTask(0); SlotEn = true; SetLedMode(2); break;
    case 2: // LTOF mode
      DispMode = 20; PixyBotActive = true;
      SetMainTask(21); break;
    case 3: // Pixy2 mode
      DispMode = 30; PixyBotActive = true;
      SetMainTask(31); break;
    case 4: // adjust settings mode
      SetSel = 0;
      DispMode = 40; PixyBotActive = false;
      SetMainTask(40); SetLedMode(0); break;
    case 99:  // select new mode
      DispMode = 99; SetMainTask(0); break;
  } DispDel = 0;  DispCnt = 0; DispNxtCnt = 0;
//  Serial.println("SetMainMode = " + String(MainMode));
}

// --------------------------------------------------------------------------------

void SetMainTask(int zMT) {
  // sets the main task and associated flags
  // immediately stop any movement
  autoOFF = 5;
  BlockAsk = 3;         // number of blocks asked for from the Pixy2 camera, default = 3
  BlockCnt = 0;         // reset the number of blocks detected by the Pixy2 camera
  DispNOP = false;      // if == true then don't update the display functions
  Drive = 0;            // reset drive direction
  DriveLast = 0;        // reset previous driving direction
  ESC = false;          // clear the ESCape flag
  Gear = 1;             // start in lowest speed range
  if (Lights) {Lamps(0);}  // true so turn white LED lamps OFF
  Lights = false;       // if == true then turn Pixy2 lights ON
  LTOFcalc = false;     // if == true then calculate speed of ranged object in m/s
  MainRun = false;      // clear MainTask inh
  MainTask = zMT;       // set MainTask
  MotorPWM(0,0,0,0);    // ensure motors are off
  PixyEn = false;       // turn OFF Pixy2 I2C comms
  Pixy_ServoReset();    // centralise Pixy2 servos, if connected
  PwmEn = false;        // if == true motor PWM is enabled
  PwmInt = 0;           // integrator used in motor control
  Range = LtofMax;      // set LTOF range to max
  RangeMax = LtofLimit; // the current limited to Range measurements
  RangeMin = LtofMin;   // the current lower limited to Range measurements
  Reset_Slots();        // reset slot counters
  SetLedMode(0);        // remove any LED flashing
  SigAp = 0;            // block array pointer to SigId data
  SigId = 0;            // object of interest signtaure ID, default is 0 (largest)
  SigTask = 0;          // object of interest recognition task pointer
  SlotEn = false;       // turn OFF slot sensor reading
  Steer = 0;            // zero steering
  SubCnt = 0;           // sub task general counter
  SubDel = 0;           // subtask delay counter
  SubTask = 0;          // reset sub task pointer
  SubTask1 = 0;         // level 1 sub task pointer
  sw0_Nop = false;      // if == true don't perform switch functions
  sw1_Nop = false;      // if == true don't perform switch functions
  if (VL53L1X_Task > 0) {VL53L1X_OFF();}  // laser disabled by default
  Z_Mode = false;       // joystick motion flag, normally false, if == true then slide mode

  // give VL53L1X 1ms to reset
  delay(1); synchLoopTimers(1);
  
  switch(MainTask) {
    case 11:  // Drive Demo 1
      break;
    case 21:  // LTOF ranging display
      VL53L1X_ON();                     // enable laser ranging task
//      if (HeadActive) {HeadTask = 1;  // activate scanning
//      } else {LTOFcalc = true;}       // calculate speed of ranged object in m/s
      break;
    case 22:  // LTOF backaway task
      VL53L1X_ON();                     // enable back away task
      PwmEn = true;                     // enable motor PWM
//      if (HeadActive) {HeadTask = 1;} // activate scanning
      break;
    case 23:  // LTOF target tracking task
      VL53L1X_ON();                     // enable target tracking task
      PwmEn = true;                     // enable motor PWM
//      if (HeadActive) {HeadTask = 1;} // activate scanning
      break;
    case 24:  // LTOF autonomous task
      VL53L1X_ON();                     // enable autonomous task
      PwmEn = true;                     // enable motor PWM
//      if (HeadActive) {HeadTask = 1;} // activate scanning
      break;
    case 31:  // Pixy2 static recognition
      if (I2C_Pixy) {PixyEn = true;}    // enable Pixy2 I2C comms
      break;
    case 32:  // Pixy2 pan & tilt recognition
      if (I2C_Pixy) {PixyEn = true;}    // enable Pixy2 I2C comms
      break;
    case 33:  // Pixy2 pan & tilt & chassis turning
      if (I2C_Pixy) {PixyEn = true;}    // enable Pixy2 I2C comms
      PwmEn = true;                     // enable motor PWM
      break;
    case 34:  // Pixy2 pan & tilt & chassis following, moderate
      if (I2C_Pixy) {PixyEn = true;}    // enable Pixy2 I2C comms
      PwmEn = true;                     // enable motor PWM
      break;
    case 35:  // Pixy2 pan & tilt & chassis following, FAST
      if (I2C_Pixy) {PixyEn = true;}    // enable Pixy2 I2C comms
      PwmEn = true;                     // enable motor PWM
      break;
    case 36:  // Pixy2 Goalie
      if (I2C_Pixy) {PixyEn = true;}    // enable Pixy2 I2C comms
      PwmEn = true;                     // enable motor PWM
      break;
  }
//  Serial.println("MainTask = " + String(MainTask));
}

// --------------------------------------------------------------------------------

void SetMaxSpeed() {
  // called when changing 'Gear' to set drive limitations
}

// --------------------------------------------------------------------------------

void Slots_ON() {
  // turn ON slot sensor readings and reset counters
  Reset_Slots(); SlotEn = true;
}

// --------------------------------------------------------------------------------

void VL53L1X_OFF() {
  // switches OFF the VL53L1X laser ranging device
  digitalWrite(XSHUT_PIN,LOW);  // shut down VL53L1X sensor
  Serial.println("VL53L1X OFF");
  RangeEn = false; VL53L1X_Task = 0;
}

// --------------------------------------------------------------------------------

void VL53L1X_ON() {
  // switches ON the VL53L1X laser ranging device
  // if it was already ON then switch it OFF then ON again
  if (RangeEn) {VL53L1X_OFF(); delay(1); synchLoopTimers(1);}
  
  digitalWrite(XSHUT_PIN,HIGH);  // enable VL53L1X sensor
  if (VL53L1X.init() == false) {
    Serial.println("VL53L1X sensor detected!"); I2C_LTOF = true;
    VL53L1X.setDistanceModeShort();         // Sets to 1.3M range
    VL53L1X.setTimingBudgetInMs(33);        // Set timing budget to 33 ms
    VL53L1X.setIntermeasurementPeriod(33);  // Set measurement period in ms
    VL53L1X.setROI(VL53L1X_ROIX,VL53L1X_ROIY,VL53L1X_OC); // Set SPAD ROI
    VL53L1X.startRanging();                 // Starts taking measurements
    Serial.println("VL53L1X ROI SPAD W= " + String(VL53L1X.getROIX()));
    Serial.println("VL53L1X ROI SPAD H= " + String(VL53L1X.getROIY()));
    RangeEn = true; VL53L1X_Task = 1;
    Range = LtofLimit; RangeLast = LtofLimit;
  } else { I2C_LTOF = false; Serial.println("VL53L1X sensor failed!"); RangeEn = false;}
}

// --------------------------------------------------------------------------------

void VL53L1X_Run() {
  // control the VL53L1X sensor using a series of tasks
  // this code gets timing information, for using the sensor more efficiently
  // once measurements start a skip timer is used to avoid unnecessary readings
  if (millis() < VL53L1X_Skip) {return;}  // allowing for reading delay 32ms
  
  switch(VL53L1X_Task) {
    // case 8 is placed at the front of the list for speed
    case 8: // wait for measurement to complete
      // this case is placed at the head of the list as it will be used most often
      // record previous range just before taking a new measurement
      RangeLast = RangeRaw;
      if (VL53L1X.checkForDataReady()) {
//        t1 = millis() - t1; // how long did the measurement take?
//        Serial.println("VL53L1X.checkForDataReady() in " + String(t1) + "ms");
//        t1 = millis();
        VL53L1X_Skip = millis() + VL53L1X_SkipDel; // set the next measurement delay
        if (VL53L1X.getRangeStatus() == 0) {
          // no ranging errors
//          t0 = micros();
          RangeRaw = VL53L1X.getDistance();
//          t0 = micros() - t0;
//          Serial.println("VL53L1X.getDistance() in " + String(t0) + "µs");
        } else {
          // if an error occurs we set the value to 2 x RangeMax
          RangeRaw = RangeMax;
        }
        LTOFt1 = millis();
//        t0 = micros();
        VL53L1X.clearInterrupt(); // 444µs clears the interrupt and data ready flag
//        t0 = micros() - t0;
//        Serial.println("VL53L1X.clearInterrupt() in " + String(t0) + "µs");
//        Serial.println("");
        // detect and remove transient low values
        if (RangeRaw < RangeMin) {RangeRaw = RangeLast;}
        // range is averaged over two measurements to reduce noise
        Range = (RangeRaw + RangeLast)/2;
        RangeUL = false;
        // limit the range to avoid erroneous values
        if (Range >= RangeMax) {Range = RangeMax; RangeUL = true;}
        RangeLL = false;
        if (Range < RangeMin) {Range = RangeMin; RangeLL = true;}
//        Serial.println("Range = " + String(Range) + "mm");
//        Serial.println(Range);
        // calculate the intermeasurement period
        LTOFperiod = LTOFt1 - LTOFt0; LTOFt0 = LTOFt1;
        // calculate the fps frequency
        LTOFfps = 1000/LTOFperiod;
        VL53L1X_Task = 8; // look for another measurement
        if (LTOFcalc) {VL53L1X_Task = 9;}
//        else {Serial.println("0,1000," + String(RangeRaw) + "," + String(Range));}
      } break;
    case 9: // speed calculation is enabled
      // due to noise on the sensor we cannot measure low speed values, and therefore
      // set s threshold of 10mm which is equivalent to 300mm/s minimum
      RangeDiff = abs(RangeRaw - RangeLast); // get the difference
      if (RangeDiff > 10) {
        // movement is > 10mm so calculate speed
        // we apply a peak detector, with 1 sec sustain window
        LTOFspd10 = (RangeDiff * 10)/LTOFperiod;
        if (LTOFspd10 > LTOFspd10L) {LTOFspd10L = LTOFspd10; LTOFspdCnt = 30;}
        else {LTOFspd10 = LTOFspd10L;}
      }
      if (LTOFspdCnt > 0) {
        LTOFspdCnt--;
        if (LTOFspdCnt < 1) {LTOFspd10 = 0; LTOFspd10L = 0;}
      }
//      Serial.println("0,1000," + String(Range) + "," + String(LTOFspd10));
      VL53L1X_Task = 8; // look for another measurement
      break;

    case 1: // switch on sensor
      digitalWrite(XSHUT_PIN,HIGH);  // enable VL53L1X sensor
      VL53L1X_Timer = micros(); 
//      Serial.println("VL53L1X XSHUT HIGH");
      VL53L1X_Task++; break;
    case 2: // check for boot, will time-out after 10ms if sensor does not respond
//      t0 = micros();
      if (VL53L1X.checkBootState()) {
//        t0 = micros() - t0;
        VL53L1X_Timer = micros() - VL53L1X_Timer;
//        Serial.println("VL53L1X.checkBootState() in " + String(t0) + "µs");
        // 825µs needed to boot the device from XSHUT going high
//        Serial.println("VL53L1X Booted! in " + String(VL53L1X_Timer) + "µs");
        VL53L1X_Task++;
      }
      else if (micros() >= (VL53L1X_Timer + 10000)) {VL53L1X_Task = 99;} // time-out
      break;
    case 3: // device booted, so initialise
//      t0 = micros();
      if (VL53L1X.init() == false) {
//        t0 = micros() - t0;
//        Serial.println("VL53L1X.init() in " + String(t0) + "µs"); // 151ms
        VL53L1X_Task++;
      } else {VL53L1X_Task = 99;} // failed to initialise, so turn sensor OFF
      break;
    case 4: // set the distance mode
//      t0 = micros();
      VL53L1X.setDistanceModeShort();         // Sets to 1.3M range
//      t0 = micros() - t0;
//      Serial.println("VL53L1X.setDistanceModeShort() in " + String(t0) + "µs");
      VL53L1X_Task++; break;
    case 5: // set the timing budget as 33ms
//      t0 = micros();
      VL53L1X.setTimingBudgetInMs(33);        // Set timing budget to 33 ms
//      t0 = micros() - t0;
//      Serial.println("VL53L1X.setTimingBudgetInMs(33) in " + String(t0) + "µs");
      VL53L1X_Task++; break;
    case 6: // set the distance mode
//      t0 = micros();
      VL53L1X.setIntermeasurementPeriod(33);  // Set measurement period in ms
//      t0 = micros() - t0;
//      Serial.println("VL53L1X.setIntermeasurementPeriod(33) in " + String(t0) + "µs");
      RangeEn = true; VL53L1X_Task++; break;
    case 7: // initiate ranging
//      t0 = micros();
      VL53L1X.startRanging();                 // 446µs Starts taking measurements
//      t0 = micros() - t0;
//      Serial.println("VL53L1X.startRanging() in " + String(t0) + "µs");
//      t1 = millis();
      VL53L1X_Skip = millis() + VL53L1X_SkipDel; // set the initial delay
      LTOFt0 = millis(); VL53L1X_Task++; break;
      
    case 99: // a time-out has occured or we want to turn the sensor OFF
      digitalWrite(XSHUT_PIN,LOW);  // disable VL53L1X sensor
//      Serial.println("VL53L1X XSHUT LOW");
      RangeEn = false; VL53L1X_Task = 0; break;
  }
}

// ---------------------------------------------------------------------------------



  //###############################################################################
  //
  //  WiFi Code
  //
  //###############################################################################

void WiFiClearRxBuff(int zNum) {
  // clears the contents of the receive buffer
  for (int zP = 0; zP < zNum;zP++) {Rx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void WiFiClearTxBuff(int zNum) {
  // clears the contents of the transmit buffer
  for (int zP = 0; zP < zNum;zP++) {Tx_Buff.ESPdata[zP] = 0;}
}

// ---------------------------------------------------------------------------------

void WiFiDisable() {
  // when either a switch is pressed or WiFi disable is required
  WiFiEn = false; WiFiCntC = 0; WiFiRestCnt = 0; C_Dn = 8; Z_Dn = 8;
}

// --------------------------------------------------------------------------------

void WiFiDisconnect() {
  // enter the WiFi disconnected state
  WiFiConnected = false;
  WiFiClearRxBuff(WiFiTx_len);
  WiFiClearTxBuff(WiFiTx_len);
  WiFiTryCnt = 0;     // reset the connection retry counter
  WiFiTx_CB = 1;      // set ready to send flag
  WiFiRxRec = false;  // clear the data received flag
  Serial.println("WiFiDisconnected!");
}

// --------------------------------------------------------------------------------

void WiFiTryToConnect() {
  // try to connect with the Wii Transceiver by sending devices name
  Tx_Buff.ESPdata[0] = 'P';
  Tx_Buff.ESPdata[1] = 'i';
  Tx_Buff.ESPdata[2] = 'x';
  Tx_Buff.ESPdata[3] = 'y';
  Tx_Buff.ESPdata[4] = 'B';
  Tx_Buff.ESPdata[5] = 'o';
  Tx_Buff.ESPdata[6] = 't';
  Tx_Buff.ESPdata[7] = '2';
  WiFiTx_len = 8;
//  t0 = micros();
  esp_err_t result = esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
  if (result == ESP_OK) {
//    Serial.println("Sent with success");
  } else {
//    Serial.println("Error sending the data");
  } 

  WiFiTryNum++; // count the total number of attempts
//  Serial.println("WiFiTryToConnect...");
}

// --------------------------------------------------------------------------------

void WiFiTxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Tx buffer of length zlen
  WiFiTx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    WiFiTx_Chk = WiFiTx_Chk ^ Tx_Buff.ESPdata[zP];
  }
}

// ---------------------------------------------------------------------------------
